"""
Simple container class for information about a link between two PDF objects.
"""
from dataclasses import dataclass, field
from typing import Any, Self

from pypdf.generic import BooleanObject, IndirectObject, NullObject, PdfObject
from yaralyzer.util.logging import log

from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError
from pdfalyzer.util.helpers.pdf_object_helper import coerce_nums_array_to_dict
from pdfalyzer.util.helpers.string_helper import bracketed, coerce_address, is_prefixed_by_any

INCOMPARABLE_PROPS = ['from_obj', 'to_obj']


@dataclass
class PdfObjectRelationship:
    from_node: 'PdfTreeNode'
    to_obj: IndirectObject
    reference_key: str
    address: str  # ints will be coerced to strings
    from_obj: Any | None = None
    is_child: bool = False
    is_indeterminate: bool = False
    is_link: bool = field(init=False)
    is_parent: bool = field(init=False)

    def __post_init__(self) -> None:
        """
        In the case of easy key/value pairs the reference_key and the address are the same but
        for more complicated references the address will be the reference_key plus sub references.
        e.g. a reference_key for a /Font labeled /F1 could be '/Resources' but the address
        might be '/Resources[/Font][/F1] if the /Font is a directly embedded reference instead of a remote one.
        """
        self.address = coerce_address(self.address)

        # Compute tree placement logic booleans
        if (has_indeterminate_prefix(self.from_node.type) and not isinstance(self.from_node.obj, (dict, list))) \
                or self.reference_key in INDETERMINATE_REF_KEYS:
            log.info(f"Indeterminate node {self.from_node} has relationship to {self.to_obj.idnum}")
            self.is_indeterminate = True

        self.is_link = self.reference_key in NON_TREE_KEYS or is_prefixed_by_any(self.from_node.label, LINK_NODE_KEYS)
        self.is_parent = self.reference_key == PARENT or (self.from_node.type == STRUCT_ELEM and self.reference_key == P)

        # TODO: there can be multiple OBJR refs to the same object... which wouldn't work w/this code
        if self.from_node.type == OBJR and self.reference_key == OBJ:
            log.info(f"Explicit (theoretically) child reference found for {OBJ} in {self.from_node}")
            self.is_child = True
        elif self.reference_key == KIDS or (self.from_node.type == STRUCT_ELEM and self.reference_key == K):
            self.is_child = True

        if self.is_parent and self.is_child:
            raise PdfWalkError(f"{self} is both parent->child and child->parent")

    @classmethod
    def build_node_references(
        cls,
        from_node: 'PdfTreeNode',
        from_obj: PdfObject | dict | None = None,
        ref_key: str | int | None = None,
        address: str | int | None = None
    ) -> list[Self]:
        """
        Recursively build list of relationships 'from_node.obj' contains referencing other PDF objects.
        Initially called with single arg from_node. Other args are use when recursable objs are scanned.

        Args:
            from_node (PdfTreeNode): PDF node that may contain references to other PDF objects
            from_obj (PdfObject | None): Raw PdfObject that may contain references to other PDF objects
            ref_key (str): The /Color, /Page etc. style string or int for ArrayObjects
            address (str): Base address for other refs, used for internal recursion into dicst and arrays
        """
        from_obj = from_node.obj if from_obj is None else from_obj
        references: list[Self] = []

        if isinstance(from_obj, IndirectObject):
            references.append(cls(from_node, from_obj, str(ref_key), address))
        elif isinstance(from_obj, list):
            for i, item in enumerate(from_obj):
                references += cls.build_node_references(from_node, item, ref_key or i, _build_address(i, address))
        elif isinstance(from_obj, dict):
            for key, val in from_obj.items():
                if key == NUMS and isinstance(val, list) and len(val) % 2 == 0:
                    val = coerce_nums_array_to_dict(val)

                references += cls.build_node_references(from_node, val, ref_key or key, _build_address(key, address))
        elif not isinstance(from_obj, (float, int, str, BooleanObject, NullObject)):
            log.debug(f"Adding no references for PdfObject reference '{ref_key}' -> '{from_obj}' ({type(from_obj).__name__})")

        # Set all returned relationships to originate from top level from_obj before returning
        for ref in references:
            ref.from_obj = from_obj

        return references

    def __eq__(self, other: Self) -> bool:
        """Note that equality does not check self.from_obj equality because we don't have the idnum"""
        for key in [k for k in vars(self).keys() if k not in INCOMPARABLE_PROPS]:
            if getattr(self, key) != getattr(other, key):
                return False

        return self.from_node.idnum == other.from_node.idnum

    def __str__(self) -> str:
        s = f"Relationship of {self.from_node} to {self.to_obj.idnum}"
        s += f" (ref_key={self.reference_key}, address='{self.address}')"
        s += f" is parent/child" if self.is_parent else ''
        s += f" is child/parent" if self.is_child else ''
        s += f" via symlink" if self.is_link else ''
        return s


def _build_address(ref_key: str | int, base_address: str | int | None = None) -> str | int:
    """
    Append either array index indicators e.g. [5] or reference_keys. reference_keys that appear in a
    PDF object are left as is. Any dict keys or array indices that refere to inner objects and not
    to the PDF object itself are bracketed.  e.g. if there's a /Width key in a /Font node,
    '/Width' is the address of the font widths. But if the /Font links to the widths through a
    /Resources dict the address will be '/Resources[/Width]'
    """
    if base_address is None:
        return ref_key
    else:
        return f"{coerce_address(base_address)}{bracketed(ref_key)}"



import re
from argparse import Namespace
from dataclasses import dataclass, field
from pathlib import Path
from typing import Self

from yaralyzer.util.helpers.shell_helper import ShellResult
from yaralyzer.util.logging import log, log_and_print, log_console

from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.util.helpers.filesystem_helper import (PDF_PARSER_PATH_ENV_VAR, PDF_PARSER_PY,
     PROJECT_ROOT, SCRIPTS_DIR, dir_str)
from pdfalyzer.util.exceptions import PdfParserError

# PDF Internal Data Regexes
CONTAINS_STREAM_REGEX = re.compile(r'\s+Contains stream$')
PDF_OBJECT_START_REGEX = re.compile(r'^obj (\d+) \d+$')

# Install info
DIDIER_STEVENS_RAW_GITHUB_URL = 'https://raw.githubusercontent.com/DidierStevens/DidierStevensSuite/master/'
PDF_PARSER_GITHUB_URL = DIDIER_STEVENS_RAW_GITHUB_URL + PDF_PARSER_PY
INSTALL_SCRIPT_PATH = SCRIPTS_DIR.joinpath('install_didier_stevens_pdf_tools.py').relative_to(PROJECT_ROOT)
PDF_PARSER_INSTALL_MSG = f"If you need to install {PDF_PARSER_PY} it's a single .py file that can be " \
                         f"found at {PDF_PARSER_GITHUB_URL}. There's a script in the Pdfalyzer repo that " \
                         f"will install it for you at {INSTALL_SCRIPT_PATH}."


@dataclass
class PdfParserManager:
    """Instances of this class manage external calls to Didier Stevens's pdf-parser.py for a given PDF."""
    path_to_pdf: Path
    output_dir: Path
    base_shell_cmd: list[str | Path] = field(init=False)
    object_ids: list[int] = field(default_factory=list)
    object_ids_containing_stream_data: list[int] = field(default_factory=list)

    @classmethod
    def from_args(cls, args: Namespace) -> Self:
        """Alternate constructor that takes the result of an `ArgumentParser.`"""
        return cls(args.file_to_scan_path, args.output_dir)

    def __post_init__(self):
        if PdfalyzerConfig.pdf_parser_path is None:
            raise PdfParserError(f"{PDF_PARSER_PATH_ENV_VAR} not configured.\n\n{PDF_PARSER_INSTALL_MSG}")

        self.base_shell_cmd = ['python', PdfalyzerConfig.pdf_parser_path, '-O', self.path_to_pdf]
        self.extract_object_ids()

    def extract_object_ids(self) -> None:
        """Examine output of pdf-parser.py to find all object IDs as well as those object IDs that have streams"""
        try:
            result = ShellResult.from_cmd(self.base_shell_cmd, verify_success=True)
            log.debug(result.output_logs(True))
        except Exception as e:
            raise PdfParserError(f"Failed to execute '{self.base_shell_cmd}' ({e})")

        current_object_id = None

        for line in result.stdout.split("\n"):
            if (match := PDF_OBJECT_START_REGEX.match(line)):
                current_object_id = int(match[1])
                self.object_ids.append(current_object_id)

            if current_object_id is None:
                continue

            if CONTAINS_STREAM_REGEX.match(line):
                log.debug(f"{current_object_id} contains a stream!")
                self.object_ids_containing_stream_data.append(current_object_id)

        log.info(f"{self.path_to_pdf} Object IDs: {self.object_ids}")
        log.info(f"{self.path_to_pdf} Objs IDs w/streams: {self.object_ids_containing_stream_data}")

    def extract_all_streams(self) -> None:
        """Use pdf-parser.py to find binary data streams in the PDF and dump each of them to a separate file"""
        output_dir_str = dir_str(self.output_dir)
        log_and_print(f"\nExtracting binary streams in '{self.path_to_pdf}' to files in '{output_dir_str}'...")
        files_written = []

        for object_id in self.object_ids_containing_stream_data:
            stream_dump_file = self.output_dir.joinpath(f'{self.path_to_pdf.name}.object_{object_id}.bin')
            shell_cmd = self.base_shell_cmd + ['-f', '-o', object_id, '-d', stream_dump_file]
            log.info(f'Dumping stream from object {object_id} with cmd:\n\n{shell_cmd}\n')

            try:
                ShellResult.from_cmd(shell_cmd, verify_success=True)
                files_written.append(stream_dump_file)
            except Exception as e:
                log.error(f"Failed to extract object ID {object_id}!")

        log_and_print(f"{len(files_written)} extracted binary streams were written to {output_dir_str}.")
        num_failures = len(self.object_ids_containing_stream_data) - len(files_written)

        if num_failures > 0:
            log_console.print(f"{num_failures} streams could not be extracted!", style='bright_red')



from dataclasses import dataclass, field
from typing import Any, Self

from pypdf.errors import PdfReadError
from pypdf.generic import DictionaryObject, IndirectObject, NullObject, NumberObject, PdfObject
from rich.text import Text

from pdfalyzer.output.highlighter import PdfHighlighter
from pdfalyzer.output.theme import (COMPLETE_THEME_DICT, DEFAULT_LABEL_STYLE, get_class_style,
     get_class_style_dim, get_class_style_italic)
from pdfalyzer.util.adobe_strings import GO_TO_E, GO_TO_R, IMAGE, S, SUBTYPE, TYPE, UNLABELED, XOBJECT
from pdfalyzer.util.helpers.pdf_object_helper import pypdf_class_name
from pdfalyzer.util.helpers.rich_helper import comma_join_txt
from pdfalyzer.util.helpers.string_helper import coerce_address, is_array_idx, props_string_indented, root_address
from pdfalyzer.util.logging import highlight, log, log_console, log_highlighter, log_trace, pdf_highlighter


@dataclass
class PdfObjectProperties:
    """
    Decorator for PyPDF PdfObject that extracts a couple of properties (type, label, etc).

    Attributes:
        obj (PdfObject): The underyling PDF object
        address (str | int): The location of the PDF object in its parent object, e.g 'Kids[0]'
        idnum (int): ID of the PDF object
        indirect_object (IndirectObject | None): IndirectObject that points to this one
        label (str): A string that meaningfully describes this object
        sub_type (str | None): The value found in the /Subtype or /S props, if it exists
        type (str | None): The value found in the /Type, defaulting to the address
    """
    obj: PdfObject
    address: str  # ints will be coerced
    idnum: int
    indirect_object: IndirectObject | None = None

    # Computed fields
    label: str = ''
    sub_type: str | None = None
    _type: str | None = None

    @property
    def type(self) -> str:
        return self._type or '???'

    @property
    def label_style(self) -> str:
        type_no_slash = self.type.removeprefix('/')
        sub_type = self.sub_type or ''

        if sub_type.startswith(GO_TO_R) or sub_type.startswith(GO_TO_E):
            return COMPLETE_THEME_DICT[PdfHighlighter.prefixed_style(GO_TO_R)]
        elif self.type == XOBJECT and sub_type == IMAGE:
            return COMPLETE_THEME_DICT[PdfHighlighter.prefixed_style(IMAGE)]

        return COMPLETE_THEME_DICT.get(PdfHighlighter.prefixed_style(type_no_slash), DEFAULT_LABEL_STYLE)

    @type.setter
    def type(self, _type: str | None):
        self._type = _type

        if self._type is None:
            log.warning(f"Unable to determine obj type for {self.idnum}, address={self.address}, obj={self.obj}!")
            self.label = f"{UNLABELED}{self.address}"
        elif self.sub_type is not None:
            self.label = f"{self.type}:{self.sub_type[1:]}"
        else:
            self.label = self._type

    def __post_init__(self,):
        self.address = coerce_address(self.address)

        if isinstance(self.obj, DictionaryObject):
            self._type = self.obj.get(TYPE)
            self.sub_type = self.obj.get(SUBTYPE) or self.obj.get(S)
        elif isinstance(self.obj, NullObject):
            self._type = NullObject.__name__

        # Use address as type if no explicit /Type, e.g. obj referenced as '/Font' is considered a /Font type.
        self.type = self._type or (root_address(self.address) if not is_array_idx(self.address) else None)
        log_trace(f"Built {repr(self)}")

    @classmethod
    def from_reference(cls, ref: IndirectObject, address: str | int) -> Self:
        """Alternate constructor to build from an IndirectObject."""
        try:
            return cls(ref.get_object(), address, ref.idnum, ref)
        except PdfReadError as e:
            log_console.print_exception()
            log.error(f"Failed to build node because, integrity not guaranteed. Error: {e}")
            return cls(ref, address, ref.idnum)

    def get_table_row(
        self,
        reference_key: str | int | None,
        pdfalyzer: 'Pdfalyzer',  # noqa: F821
        empty_3rd_col: bool = False
    ) -> tuple[Text, Text, Text]:
        """Extract property at `reference_key` and build a formatted 3-tuple for use in Rich tables."""
        if reference_key is None:
            key_style = ''
            row_obj = self.obj
        else:
            if isinstance(self.obj, dict):
                row_obj = self.obj.get(reference_key)  # NOTE: self.obj[k] turns IndirectObjects into PdfObjects!
            elif isinstance(self.obj, list) and isinstance(reference_key, int):
                row_obj = self.obj[reference_key]
            else:
                raise Exception(f"Invalid ref key/obj combo! ref_key={reference_key}, obj={repr(self)}")

            if isinstance(reference_key, int):
                key_style = 'grey'
            else:
                key_style = pdf_highlighter.get_style(reference_key) or log_highlighter.get_style(reference_key)

        with_resolved_refs = self._resolve_references(reference_key, row_obj, pdfalyzer)
        value_style = key_style if isinstance(row_obj, str) else get_class_style(row_obj)
        col1 = Text(f"{reference_key}", style=key_style)
        # Prefix the Text() with empty string to set unstyled chars to style of the object they are in.
        col2 = Text('', style=value_style).append_text(self._obj_to_rich_text(with_resolved_refs))
        col3 = Text('' if empty_3rd_col else pypdf_class_name(row_obj), style=get_class_style_dim(row_obj))
        return (col1, col2, col3)

    def node_label(self, underline: bool = True) -> Text:
        """Colored text representation of a PDF node. Example: <5:FontDescriptor(Dictionary)>."""
        text = Text('<', style='white')
        text.append(f'{self.idnum}', style='bright_white')
        text.append(':', style='white')
        text.append(self.label[1:], style=f"bold {self.label_style} {'underline' if underline else ''}")
        text.append('(', style='white')
        text.append(pypdf_class_name(self.obj), style=get_class_style_italic(self.obj))
        text.append(')', style='white')
        text.append('>')
        return text

    def _resolve_references(self, reference_key: str | int, obj: PdfObject, pdfalyzer: 'Pdfalyzer') -> Any:  # noqa: F821
        """Recursively build the same data structure except IndirectObjects are resolved to nodes."""
        if isinstance(obj, NumberObject):
            return obj.as_numeric()
        elif isinstance(obj, IndirectObject):
            node = pdfalyzer.find_node_by_idnum(obj.idnum)
            return node.pdf_object if node else self.from_reference(obj, reference_key)
        elif isinstance(obj, dict):
            return {k: self._resolve_references(k, v, pdfalyzer) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._resolve_references(reference_key, item, pdfalyzer) for item in obj]
        else:
            return obj

    # TODO: this doesn't recurse?
    @classmethod
    def _obj_to_rich_text(cls, obj: Any) -> Text:
        """Recurse through `obj` and build a `Text` object."""
        if isinstance(obj, dict):
            key_value_pairs = [Text(f"{k}: ").append_text(cls._obj_to_rich_text(v)) for k, v in obj.items()]
            return Text('{').append_text(comma_join_txt(key_value_pairs)).append('}')
        elif isinstance(obj, list):
            items = [cls._obj_to_rich_text(item) for item in obj]
            return Text('[').append_text(comma_join_txt(items)).append(']')
        else:
            return cls._to_text(obj)

    @classmethod
    def _to_text(cls, obj: Any) -> Text:
        """Handles styling non-recursable objects."""
        if isinstance(obj, cls):
            return cls.__rich_without_underline__(obj)
        elif isinstance(obj, str):
            if 'http' in obj or obj.startswith('/'):
                return highlight(obj)
            else:
                return Text(obj)
        else:
            return Text(str(obj), style=get_class_style(obj))

    def __repr__(self) -> str:
        return f"{type(self).__name__}(" + props_string_indented(self) + '\n)'

    def __rich_without_underline__(self) -> Text:
        return self.node_label(underline=False)

    def __rich__(self) -> Text:
        return self.node_label()

    def __str__(self) -> str:
        return self.__rich__().plain



from argparse import Namespace
from dataclasses import dataclass
from functools import partial, update_wrapper
from typing import Callable, Self

from yaralyzer.util.logging import log

# Analysis selection sections
DOCINFO = 'docinfo'
TREE = 'tree'
RICH = 'rich'
FONTS = 'fonts'
COUNTS = 'counts'
YARA = 'yara'
STREAMS = 'streams'

DEFAULT_SECTIONS = [DOCINFO, TREE, RICH, FONTS, COUNTS, YARA]
ALL_SECTIONS = DEFAULT_SECTIONS + [STREAMS]

ALL_STREAMS = -1


@dataclass
class OutputSection:
    """Helper class to keep track of which analyses are requested by the user."""
    argument: str
    method: Callable

    @classmethod
    def selected_sections(cls, args: Namespace, presenter: 'PdfalyzerPresenter') -> list[Self]:  # noqa: F821
        """
        Determine which of the tree visualizations, font scans, etc should be run.
        If nothing is specified output ALL sections other than --streams which is v. slow/verbose.

        Args:
            args: parsed command line arguments
            pdfalyzer: the `pdfalyzer` instance whose methods will be called to produce output

        Returns:
            list[OutputSection]: List of `OutputSection` namedtuples with 'argument' and 'method' fields
        """
        # Create a partial for print_font_info() because it's the only one that can take an argument
        # partials have no __name__ so update_wrapper() propagates the 'print_font_info' as this partial's name
        stream_id = None if args.streams == ALL_STREAMS else args.streams
        stream_scan = partial(presenter.print_streams_analysis, idnum=stream_id)
        update_wrapper(stream_scan, presenter.print_streams_analysis)

        # 1st element string matches the argument in 'select' group
        # 2nd is fxn to call if selected.
        # Top to bottom is the default order of output.
        possible_output_sections = [
            cls(DOCINFO, presenter.print_document_info),
            cls(TREE, presenter.print_tree),
            cls(RICH, presenter.print_rich_table_tree),
            cls(FONTS, presenter.print_font_info),
            cls(COUNTS, presenter.print_summary),
            cls(YARA, presenter.print_yara_results),
            cls(STREAMS, stream_scan),
        ]

        output_sections = [section for section in possible_output_sections if vars(args)[section.argument]]

        if len(output_sections) == 0:
            log.warning("No output section specified so outputting all sections except --streams...")
            return [section for section in possible_output_sections if section.argument != STREAMS]
        else:
            return output_sections

    @staticmethod
    def all_sections_chosen(args: Namespace) -> bool:
        """Returns True if all flags are set or no flags are set."""
        return len([s for s in ALL_SECTIONS if vars(args)[s]]) == len(ALL_SECTIONS)



from yaralyzer.output.console import console
from yaralyzer.util.logging import log


def ocr_text(image: "Image.Image", image_name: str) -> str | None:  # noqa F821
    """Use pytesseract to OCR the text in the image and return it as a string."""
    import pytesseract
    from PIL import Image  # noqa: F401
    text = None

    try:
        text = pytesseract.image_to_string(image)
    except pytesseract.pytesseract.TesseractError:
        console.print_exception()
        log.warning(f"Tesseract OCR failure '{image_name}'! No OCR text extracted...")
    except OSError as e:
        if 'truncated' in str(e):
            log.warning(f"Truncated image file '{image_name}'!")
        else:
            console.print_exception()
            log.error(f"Error while extracting '{image_name}'!")
            raise e
    except Exception as e:
        console.print_exception()
        log.error(f"Error while extracting '{image_name}'!")
        raise e

    return None if text is None else text.strip()



"""
PDFalyzer: Analyze and explore the structure of PDF files.
"""
import time
from dataclasses import dataclass, field
from io import BufferedReader
from pathlib import Path
from typing import Iterator

from anytree import LevelOrderIter, SymlinkNode
from anytree.search import findall, findall_by_attr
from pypdf import PdfReader
from pypdf.errors import DependencyError, FileNotDecryptedError, PdfReadError
from pypdf.generic import DictionaryObject, IndirectObject
from rich.prompt import Prompt
from rich.text import Text
from yaralyzer.output.console import console
from yaralyzer.output.file_hashes_table import BytesInfo
from yaralyzer.util.exceptions import print_fatal_error, print_fatal_error_and_exit

from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.decorators.document_model_printer import highlighted_raw_pdf_obj_str
from pdfalyzer.decorators.indeterminate_node import IndeterminateNode
from pdfalyzer.decorators.pdf_tree_node import PdfTreeNode
from pdfalyzer.decorators.pdf_tree_verifier import PdfTreeVerifier
from pdfalyzer.font_info import FontInfo
from pdfalyzer.util.helpers.pdf_object_helper import RefAndObj, describe_obj
from pdfalyzer.pdf_object_relationship import PdfObjectRelationship
from pdfalyzer.util import adobe_strings
from pdfalyzer.util.argument_parser import is_pdfalyze_script
from pdfalyzer.util.exceptions import PdfParserError, PdfWalkError
from pdfalyzer.util.logging import log, log_trace  # Triggers log setup
from pdfalyzer.util.pdf_parser_manager import PdfParserManager

MISSING_NODE_WARN_THRESHOLD = 200
NODE_COUNT_WARN_THRESHOLD = 10_000
PASSWORD_PROMPT = Text(f"\nThis PDF is encrypted. What's the password?", style='bright_cyan bold')
TRAILER_FALLBACK_ID = 10_000_000
PYPDF_ERROR_MSG = "Failed to open file with PyPDF. Consider filing a PyPDF bug report: https://github.com/py-pdf/pypdf/issues"


@dataclass
class Pdfalyzer:
    """
    Walks a PDF's internals and builds the PDF logical structure tree.

    Each of the PDF's internal objects is wrapped in a `PdfTreeNode` object. The tree is managed
    by the `anytree` library. Information about the tree as a whole is stored in this class.
    Once the PDF is parsed this class provides access to info about or from the underlying PDF tree.

    Args:
        pdf_path (str): Path to the PDF file
        password (str | None): Password used to decrypt the PDF (if it's encrypted)

    Attributes:
        font_infos (list[FontInfo]): Font summary objects
        font_info_extraction_error (Exception | None): Error encountered extracting FontInfo (if any)
        max_generation (int): Max revision number ("generation") encounted in this PDF.
        nodes_encountered (dict[int, PdfTreeNode]): Nodes we've traversed already even if not in tree yet.
        pdf_bytes (bytes): PDF binary data.
        pdf_bytes_info (BytesInfo): File size, hashes, and other data points about the PDF's raw bytes.
        pdf_filehandle (BufferedReader): File handle that reads the PDF.
        pdf_size (int): Number of nodes as extracted from the PDF's Trailer node.
        pdf_tree (PdfTreeNode): The top node of the PDF data structure tree.
        verifier (PdfTreeVerifier): PdfTreeVerifier that can validate the PDF has been walked successfully.
        _indeterminate_ids (set[int]): See INDETERMINATE_REF_KEYS comment
        _tree_nodes (dict[int, PdfTreeNode): ID cache for nodes that are in the tree
    """
    pdf_path: Path
    password: str | None = None

    # Non-arguments:
    font_infos: list[FontInfo] = field(default_factory=list)
    font_info_extraction_error: Exception | None = None
    idnums_found_by_parser: list[int] | None = None
    max_generation: int = 0
    nodes_encountered: dict[int, PdfTreeNode] = field(default_factory=dict)
    num_nodes: int | None = None
    pdf_bytes: bytes = field(init=False)
    pdf_bytes_info: BytesInfo = field(init=False)
    pdf_filehandle: BufferedReader = field(init=False)
    pdf_reader: PdfReader = field(init=False)
    pdf_tree: PdfTreeNode = field(init=False)
    verifier: PdfTreeVerifier = field(init=False)
    _indeterminate_ids: set[int] = field(default_factory=set)
    _tree_nodes: dict[int, PdfTreeNode | None] = field(default_factory=dict)

    @property
    def pdf_basename(self):
        return self.pdf_path.name

    def __post_init__(self):
        self.pdf_path = Path(self.pdf_path)
        started_at = time.perf_counter()

        try:
            self.pdf_filehandle = open(self.pdf_path, 'rb')  # Filehandle must stay open so PyPDF can perform seeks
            self.pdf_reader = PdfReader(self.pdf_filehandle)
        except DependencyError as e:
            self._handle_fatal_error(f"Missing dependency required for this file.", e)
        except FileNotFoundError as e:
            self._handle_fatal_error(f"Invalid file", e)
        except PdfReadError as e:
            self._handle_fatal_error(f'PdfReadError: "{self.pdf_path}" doesn\'t seem to be a valid PDF file.', e)
        except Exception as e:
            console.print_exception()
            self._handle_fatal_error(f"{PYPDF_ERROR_MSG}", e)

        if self.pdf_reader.is_encrypted:
            if not self.pdf_reader.decrypt(self.password or Prompt.ask(PASSWORD_PROMPT)):
                self._handle_fatal_error(f"Wrong password", FileNotDecryptedError("encrypted PDF"))

        # Load bytes etc
        self.pdf_bytes = self.pdf_path.read_bytes()
        self.pdf_bytes_info = BytesInfo(self.pdf_bytes)

        # Bootstrap the root of the tree with the trailer. PDFs are always read trailer first.
        # Technically the trailer has no PDF Object ID but we set it to the /Size of the PDF.
        trailer = self.pdf_reader.trailer
        self.num_nodes = trailer.get(adobe_strings.SIZE)
        trailer_id = self.num_nodes if self.num_nodes is not None else TRAILER_FALLBACK_ID
        self.pdf_tree = PdfTreeNode.from_obj(trailer, adobe_strings.TRAILER, trailer_id)
        self.nodes_encountered[self.pdf_tree.idnum] = self.pdf_tree

        if not self.num_nodes:
            log.warning(f"Could not determine number of nodes in this PDF!")
        elif self.num_nodes > NODE_COUNT_WARN_THRESHOLD:
            log.warning(f"This PDF has {self.num_nodes:,} nodes; could take a while to parse...")

        # Build tree by recursively following relationships between nodes
        self.walk_node(self.pdf_tree)

        # After scanning all objects we place nodes whose position was uncertain, extract fonts, and verify
        self._resolve_indeterminate_nodes()
        self._resolve_missing_nodes()
        self._extract_font_infos()
        self.verifier = PdfTreeVerifier(self)

        # Create SymlinkNodes for relationships between PDF objects that are not parent/child relationships.
        # (Do this last because it has the side effect of making a lot more nodes)
        for node in self.node_iterator():
            if not isinstance(node, SymlinkNode):
                node.symlink_non_tree_relationships()

        log.info(f"PDF walk completed in {time.perf_counter() - started_at:.2} seconds.")

    def close(self) -> None:
        """Close any open filehandles."""
        for attr in ['pdf_reader', 'pdf_filehandle']:
            if attr in vars(self):
                getattr(self, attr).close()

    def find_node_by_idnum(self, idnum: int) -> PdfTreeNode | None:
        """Find node with `idnum` in the tree. Return `None` if that node is not reachable from the root."""
        self._tree_nodes[idnum] = self._tree_nodes.get(idnum) or self.find_node_with_attr('idnum', idnum, True)
        return self._tree_nodes[idnum]

    def find_node_with_attr(self, attr: str, value: str | int, raise_if_multiple: bool = False) -> PdfTreeNode | None:
        """Find node with a property where you only expect one of those nodes to exist."""
        nodes = self.find_nodes_with_attr(attr, value)

        if len(nodes) == 0:
            return None
        elif len(nodes) > 1:
            msg = f"Found {len(nodes)} nodes with {attr}={value}, expected 1! {nodes}"

            if raise_if_multiple:
                raise PdfWalkError(msg)
            else:
                log.warning(msg)

        return nodes[0]

    def find_nodes_with_attr(self, attr_name: str, attr_value: str | int) -> list[PdfTreeNode]:
        """Find nodes in tree where 'attr_name' prop has the value 'attr_value'."""
        return [
            node for node in findall_by_attr(self.pdf_tree, name=attr_name, value=attr_value)
            if not isinstance(node, SymlinkNode)
        ]

    def is_in_tree(self, search_for_node: PdfTreeNode) -> bool:
        """Returns true if `search_for_node` is in the tree already."""
        return bool(self.find_node_by_idnum(search_for_node.idnum))

    def missing_node_ids(self) -> list[int]:
        """We expect to see all ordinals up to the number of nodes /Trailer claims exist as obj IDs."""
        all_object_ids = []

        # Try to get the list of object IDs with pdf-parser.py
        if PdfalyzerConfig.pdf_parser_path:
            if self.idnums_found_by_parser is None:
                try:
                    self.idnums_found_by_parser = PdfParserManager(self.pdf_path, PdfalyzerConfig.args.output_dir).object_ids
                    all_object_ids = self.idnums_found_by_parser
                    log.info(f"pdf-parser.py found {len(all_object_ids)} object IDs to verify...")

                    if self.num_nodes and self.num_nodes != (len(self.idnums_found_by_parser) + 1):
                        log.warning(f"pdf-parser.py found {len(self.idnums_found_by_parser)} objs but PDF reports {self.num_nodes}!")
                except Exception as e:
                    log.warning(f"Failed to extract object IDs with pdf-parser.py")
                    self.idnums_found_by_parser = []
            else:
                all_object_ids = self.idnums_found_by_parser

        # Fall back to all IDs between 0 and self.num_nodes
        if not all_object_ids:
            if self.num_nodes is None:
                log.error(f"no pdf-parser.py and {adobe_strings.SIZE} not found in PDF trailer; cannot verify all nodes are in tree")
                return []

            all_object_ids = [i for i in range(1, self.num_nodes)]

        return [i for i in all_object_ids if self.find_node_by_idnum(i) is None]

    def node_iterator(self) -> Iterator[PdfTreeNode]:
        """Iterate over nodes walking the tree from the top, grouped by distance from the root."""
        return LevelOrderIter(self.pdf_tree)

    def nodes_without_parents(self) -> list[PdfTreeNode]:
        """Return nodes that were encountered but have no parent set yet."""
        return [n for n in self.unplaced_encountered_nodes() if n.parent is None]

    def ref_and_obj_for_id(self, idnum: int) -> RefAndObj:
        """Build a new IndirectObject and PdfObject based on what's in the PDF."""
        ref = IndirectObject(idnum, self.max_generation, self.pdf_reader)

        try:
            obj = ref.get_object()
        except PdfReadError as e:
            if 'Invalid Elementary Object' in str(e):
                log.error(f"pypdf failed to find bad object: {e}")
                obj = None
            else:
                console.print_exception()
                log.error(str(e))
                raise e

        return RefAndObj(ref, obj)

    def stream_nodes(self) -> list[PdfTreeNode]:
        """List of actual nodes (not SymlinkNodes) containing streams sorted by PDF object ID"""
        stream_filter = lambda node: node.contains_stream() and not isinstance(node, SymlinkNode)  # noqa: E731
        return sorted(findall(self.pdf_tree, stream_filter), key=lambda r: r.idnum)

    def unplaced_encountered_nodes(self) -> list[PdfTreeNode]:
        """Nodes that were encountered by walk_node() but didn't end up in the tree."""
        return [node for id, node in self.nodes_encountered.items() if self.find_node_by_idnum(id) is None]

    def walk_node(self, node: PdfTreeNode) -> None:
        """Recursively walk the PDF's tree structure starting at a given node."""
        log.info(f'walk_node() called with {node}. Object dump:\n{highlighted_raw_pdf_obj_str(node.obj, node.label)}')
        nodes_to_walk_next = [self._add_relationship_to_pdf_tree(r) for r in node.references_to_other_nodes()]
        node.all_references_processed = True

        for next_node in [n for n in nodes_to_walk_next if not (n is None or n.all_references_processed)]:
            if not next_node.all_references_processed:
                self.walk_node(next_node)

            self.find_node_by_idnum(next_node.idnum)  # Trigger update of self._tree_nodes cache

    def _add_relationship_to_pdf_tree(self, relationship: PdfObjectRelationship) -> PdfTreeNode | None:
        """
        Place the `relationship` node in the tree. Returns an optional node that should be
        placed in the PDF node processing queue.
        """
        log.info(f'Assessing {relationship}...')
        was_seen_before = (relationship.to_obj.idnum in self.nodes_encountered)  # Must come before _build_or_find()
        from_node = relationship.from_node
        to_node = self._build_or_find_node(relationship.to_obj, relationship.address)
        self.max_generation = max([self.max_generation, relationship.to_obj.generation or 0])

        # If one is already a parent/child of the other there's nothing to do
        if to_node == from_node.parent or to_node in from_node.children:
            log.debug(f"  {from_node} and {to_node} are already parent/child")
            return None

        # NOTE: Many branches return None
        if relationship.is_parent or relationship.is_child:
            # If there's an explicit /Parent or /Kids relationship then we know the correct relationship
            log.debug(f"  Explicit parent/child link: {relationship}")

            if relationship.is_parent:
                from_node.set_parent(to_node)
            elif to_node.parent is not None:
                log.info(f"{relationship} fail: {to_node} parent is already {to_node.parent}")
            else:
                from_node.add_child(to_node)

            # Remove this to_node from inteterminacy now that it's got a child or parent
            if relationship.to_obj.idnum in self._indeterminate_ids:
                log.info(f"  Found {relationship} => {to_node} was marked indeterminate but now placed")
                self._indeterminate_ids.remove(relationship.to_obj.idnum)
        elif relationship.is_indeterminate or relationship.is_link or was_seen_before:
            # If the relationship is indeterminate or we've seen the PDF object before, add it as
            # a non-tree relationship for now. An attempt to place the node will be made at the end.
            to_node.add_non_tree_relationship(relationship)

            # If we already encountered 'to_node' then skip adding it to the queue of nodes to walk
            if was_seen_before:
                if relationship.to_obj.idnum not in self._indeterminate_ids and to_node.parent is None:
                    raise PdfWalkError(f"{relationship} - ref has no parent and is not indeterminate")
                else:
                    log_trace(f"  Already saw {relationship}; not scanning next")
                    return None
            elif relationship.is_indeterminate or (relationship.is_link and not self.is_in_tree(to_node)):
                # Indeterminate relationships need to wait until everything has been scanned to be placed
                log.info(f'  Indeterminate ref {relationship}')
                self._indeterminate_ids.add(to_node.idnum)
            elif relationship.is_link:
                # Link nodes like /Dest are usually just links between nodes
                log.debug(f"  Link ref {relationship}")
        else:
            # If no other conditions are met make from_node the parent of to_node
            from_node.add_child(to_node)

        # /StructElems in a /StructTreeRoot hierarchy sometimes have no /Type so we set it manually
        if to_node.type == adobe_strings.K and adobe_strings.STRUCT_TREE_ROOT in to_node.tree_address():
            to_node.pdf_object.type = adobe_strings.STRUCT_ELEM

        return to_node

    def _catalog_node(self) -> PdfTreeNode | None:
        return self.find_node_with_attr('type', adobe_strings.CATALOG)

    def _handle_fatal_error(self, msg: str, e: Exception) -> None:
        self.close()

        # Only exit if running in a 'pdfalyze some_file.pdf context', otherwise raise Exception.
        if is_pdfalyze_script:
            print_fatal_error_and_exit(msg, e)
        else:
            print_fatal_error(msg, e)
            raise e

    def _info_node(self) -> PdfTreeNode | None:
        return self.find_node_with_attr('type', adobe_strings.INFO)

    def _resolve_indeterminate_nodes(self) -> None:
        """Place indeterminate nodes in the tree."""
        indeterminate_nodes = [self.nodes_encountered[idnum] for idnum in self._indeterminate_ids]
        indeterminate_nodes_string = "\n   ".join([f"{node}" for node in indeterminate_nodes])
        log.info(f"Resolving {len(indeterminate_nodes)} indeterminate nodes: {indeterminate_nodes_string}")

        for node in indeterminate_nodes:
            if node.parent is not None:
                log.info(f"{node} marked indeterminate but has parent: {node.parent}")
                continue

            IndeterminateNode(node).place_node()

    def _resolve_missing_nodes(self) -> None:
        """Make a best effort to place nodes we have so far failed to get into the tree hierarchy."""
        missing_node_ids = self.missing_node_ids()

        if len(missing_node_ids) > MISSING_NODE_WARN_THRESHOLD:
            log.warning(f"Found {len(missing_node_ids)} missing node IDs. This could take a while to sort out...")

        # Place /ObjStm at root if no other location found.
        for idnum in missing_node_ids:
            ref_and_obj = self.ref_and_obj_for_id(idnum)
            ref = ref_and_obj.ref
            obj = ref_and_obj.obj
            # Make sure we didn't already fix this node up in the course of other repairs
            node = self.find_node_by_idnum(ref.idnum)

            if not isinstance(obj, DictionaryObject) or (node is not None and node.parent):
                continue

            # Handle special Linearization info nodes
            if obj.get(adobe_strings.TYPE) is None and '/Linearized' in obj:
                log.warning(f"Placing special /Linearized node {describe_obj(ref_and_obj)} as child of root")
                self.pdf_tree.add_child(self._build_or_find_node(ref, '/Linearized'))
            elif obj.get(adobe_strings.TYPE) == adobe_strings.OBJ_STM:
                log.warning(f"Forcing homeless {describe_obj(ref_and_obj)} to appear as child of root node")
                self.pdf_tree.add_child(self._build_or_find_node(ref, adobe_strings.OBJ_STM))
            elif obj.get(adobe_strings.TYPE) == adobe_strings.XOBJECT and obj.get(adobe_strings.SUBTYPE) == '/Form':
                if (form := self.find_node_with_attr('type', adobe_strings.ACRO_FORM)):
                    log.warning(f"Forcing homeless {describe_obj(ref_and_obj)} to be child of {adobe_strings.ACRO_FORM}")
                    form.add_child(self._build_or_find_node(ref, adobe_strings.XOBJECT))

        # Force /Pages to be children of /Catalog
        for node in self.nodes_without_parents():
            if node.type == adobe_strings.PAGES and (catalog_node := self._catalog_node()):
                log.warning(f"Forcing orphaned {adobe_strings.PAGES} node {node} to be child of {catalog_node}")
                node.set_parent(catalog_node)

    def _extract_font_infos(self) -> None:
        """Extract information about fonts in the tree and place it in `self.font_infos`."""
        for node in self.node_iterator():
            if not (isinstance(node.obj, dict) and adobe_strings.RESOURCES in node.obj):
                continue

            log.debug(f"Extracting fonts from node with '{adobe_strings.RESOURCES}' key: {node}...")
            known_font_ids = [fi.idnum for fi in self.font_infos]

            try:
                self.font_infos += [
                    fi for fi in FontInfo.extract_font_infos(node.obj)
                    if fi.idnum not in known_font_ids
                ]
            except Exception as e:
                self.font_info_extraction_error = e
                console.line()
                log.warning(f"Failed to extract font information from node: {node} (error: {e})")
                console.line()

        self.font_infos = sorted(self.font_infos, key=lambda fi: fi.idnum)

    def _build_or_find_node(self, relationship: IndirectObject, relationship_key: str) -> PdfTreeNode:
        """If node in self.nodes_encountered already then return it, otherwise build a node and store it."""
        if relationship.idnum in self.nodes_encountered:
            return self.nodes_encountered[relationship.idnum]

        log_trace(f"Building node for {relationship}")
        new_node = PdfTreeNode.from_reference(relationship, relationship_key)
        self.nodes_encountered[relationship.idnum] = new_node
        return new_node

    def _print_nodes_encountered(self) -> None:
        """Debug method that displays which nodes have already been walked."""
        for i in sorted(self.nodes_encountered.keys()):
            console.print(f'{i}: {self.nodes_encountered[i]}')



import code
import sys
from argparse import Namespace

from pypdf import PdfWriter
from pypdf.errors import PdfReadError
from rich.text import Text
from yaralyzer.output.console import console
from yaralyzer.output.file_export import invoke_rich_export
from yaralyzer.util.exceptions import print_fatal_error
from yaralyzer.util.logging import invocation_txt, log_console

from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.decorators.pdf_file import PdfFile
from pdfalyzer.output.pdfalyzer_presenter import PdfalyzerPresenter
from pdfalyzer.pdfalyzer import Pdfalyzer
from pdfalyzer.util.argument_parser import parser
from pdfalyzer.util.cli_tools_argument_parsers import (MAX_QUALITY, parse_combine_pdfs_args,
     parse_pdf_page_extraction_args, parse_text_extraction_args)
from pdfalyzer.util.exceptions import PdfParserError
from pdfalyzer.util.helpers.filesystem_helper import file_size_in_mb, set_max_open_files
from pdfalyzer.util.helpers.interaction_helper import ask_to_proceed
from pdfalyzer.util.logging import log  # Trigger log setup
from pdfalyzer.util.output_section import OutputSection
from pdfalyzer.util.pdf_parser_manager import PdfParserManager

PdfalyzerConfig.init(parser)


def pdfalyze():
    """Main entry point for The Pdfalyzer command line tool."""
    args = PdfalyzerConfig.parse_args()
    pdfalyzer = Pdfalyzer(args.file_to_scan_path, args.password)
    presenter = PdfalyzerPresenter(pdfalyzer)

    # Binary stream extraction is a special case
    if args.extract_binary_streams:
        try:
            PdfParserManager.from_args(args).extract_all_streams()
        except PdfParserError as e:
            print_fatal_error('Failed to extract binary streams!', e)

        sys.exit()

    # The method that gets called is related to the argument name. See 'possible_output_sections' list in
    # argument_parser.py. Analysis exports wrap themselves around the methods that actually generate the analyses.
    for section in OutputSection.selected_sections(args, presenter):
        args._export_basepath = PdfalyzerConfig.get_export_basepath(section.method)

        if args.echo_command:
            console.print(invocation_txt())

        section.method()

        if args.export_txt:
            invoke_rich_export(console.save_text, args)
        if args.export_html:
            invoke_rich_export(console.save_html, args)
        if args.export_svg:
            invoke_rich_export(console.save_svg, args)

        # Clear the buffer (if we have one) to make way for next analysis section to be printed into it
        del console._record_buffer[:]

    # Drop into interactive shell if requested
    if args.interact:
        code.interact(local=locals())

    # Non-zero error code if PDF was not verified (must be checked before closing file handle)
    exit_code = int(not bool(pdfalyzer.verifier.was_successful() or args.allow_missed_nodes))
    pdfalyzer.verifier.log_missing_node_warnings()
    pdfalyzer.close()
    exit(exit_code)


def combine_pdfs():
    """
    Script method to combine multiple PDFs into one. Invocable with 'combine_pdfs PDF1 [PDF2...]'.
    Example: https://github.com/py-pdf/pypdf/blob/main/docs/user/merging-pdfs.md
    """
    args = parse_combine_pdfs_args()
    set_max_open_files(args.number_of_pdfs)
    merger = PdfWriter()

    for pdf in args.pdfs:
        try:
            log_console.print(f"  -> Merging '{pdf}'...", style='dim')
            merger.append(pdf)
        except PdfReadError as e:
            log_console.print(f"      -> Failed to merge '{pdf}'! {e}", style='red')
            ask_to_proceed()

    # Iterate through pages and compress, lowering image quality if requested
    # See https://pypdf.readthedocs.io/en/latest/user/file-size.html#reducing-image-quality
    for i, page in enumerate(merger.pages):
        if args.image_quality < MAX_QUALITY:
            for j, img in enumerate(page.images):
                log_console.print(
                    f"  -> Reducing image #{j + 1} quality on page {i + 1} to {args.image_quality}...",
                    style='dim'
                )

                img.replace(img.image, quality=args.image_quality)

        log_console.print(f"  -> Compressing page {i + 1}...", style='dim')
        page.compress_content_streams()  # This is CPU intensive!

    log_console.print(f"\nWriting '{args.output_file}'...")
    merger.compress_identical_objects(remove_identicals=True, remove_orphans=True)
    merger.write(args.output_file)
    merger.close()
    txt = Text('').append(f"  -> Wrote ")
    txt.append(str(file_size_in_mb(args.output_file)), style='cyan').append(" megabytes\n")
    log_console.print(txt)


def extract_pdf_pages() -> None:
    """Extract a range of pages from a PDF to a new PDF."""
    args = parse_pdf_page_extraction_args()
    PdfFile(args.pdf_file).extract_page_range(args.page_range, destination_dir=args.destination_dir)


def extract_pdf_text() -> None:
    """Extract text from a list of file or from all PDF files in a list of directories."""
    args: Namespace = parse_text_extraction_args()
    console.line()

    for file_path in args.files_to_process:
        PdfFile(file_path).print_extracted_text(args.page_range, args.print_as_parsed)
        console.line(2)



"""
Log formatting and redirection. This file yields a lot of pypdf warnings:
    pdfalyze ../pypdf/tests/pdf_cache/iss2812.pdf
"""
import logging

import pypdf   # noqa: F401  # needed to trigger pypdf logger setup?
from rich.logging import RichHandler
from rich.text import Text
# Other files could import yaralyzer's log directly but they do it from here to trigger logging setup
from yaralyzer.util.logging import DEFAULT_LOG_HANDLER_KWARGS, log, log_console, log_trace

from pdfalyzer.output.highlighter import PYPDF_LOG_PFX_PATTERN, log_highlighter, pdf_highlighter

PYPDF_LOG_PFX = PYPDF_LOG_PFX_PATTERN.replace("\\", '')


def highlight(text: str | Text) -> Text:
    return pdf_highlighter(log_highlighter(text))


# Common RichHandler kwargs
log_handler_kwargs = {'highlighter': log_highlighter, **DEFAULT_LOG_HANDLER_KWARGS}

# Redirect pypdf logs and prefix them with '(pypdf)'
pypdf_log_handler = RichHandler(**log_handler_kwargs)
pypdf_log_handler.setLevel(logging.WARNING)
pypdf_log_handler.formatter = logging.Formatter(PYPDF_LOG_PFX + ' %(message)s')
logging.getLogger("pypdf").addHandler(pypdf_log_handler)

# pdfalyzer log highlighting
# TODO: this probably removes the LOG_DIR file write handler?
log.handlers = [RichHandler(**log_handler_kwargs)]



"""
Functions for miscellaneous Rich text/string pretty printing operations.
"""
from rich.panel import Panel
from rich.padding import Padding
from rich.text import Text
from yaralyzer.output.console import console

ARROW_BULLET = '➤ '


def attention_getting_panel(text: Text, title: str, style: str = 'white on red') -> Padding:
    p = Panel(text, padding=(2), title=title, style=style)
    return Padding(p, pad=(1, 10, 2, 10))


def bullet_text(msg: str | Text, style: str = '') -> Text:
    if isinstance(msg, str):
        msg = Text(msg, style=style)

    return Text(ARROW_BULLET).append(msg)


def comma_join_txt(text_objs: list[Text]) -> Text:
    return Text(", ").join(text_objs)


def error_text(text: str | Text) -> Text:
    msg = Text('').append(f"ERROR", style='bright_red').append(": ")

    if isinstance(text, Text):
        return msg + text
    else:
        return msg.append(text)


def indented_bullet(msg: str | Text, style: str = '') -> Text:
    return Text('  ') + bullet_text(msg, style)


def indent_padding(indent: int) -> tuple[int, int, int, int]:
    return (0, 0, 0, indent)


def mild_warning(msg: str) -> None:
    console.print(indented_bullet(Text(msg, style='color(228) dim')))


def number_and_pct(_number: int, total: int, digits: int = 1) -> Text:
    """Return e.g. '8 (80%)'."""
    return Text(str(_number), style='bright_white').append_text(pct_txt(_number, total, digits))


def vertically_padded_panel(title, **kwargs) -> Padding:
    # kwargs['expand'] = False if kwargs.get('expand') is None else kwargs['expand']
    kwargs['style'] = kwargs.get('style', 'reverse')
    kwargs['width'] = kwargs.get('width', 70)
    return Padding(Panel(title, **kwargs), (2, 0, 1, 0))


def pct_txt(_number: int, total: int, digits: int = 1) -> Text:
    """Return nicely formatted percentage, e.g. '(80%)'."""
    pct = (100 * float(_number) / float(total)).__round__(digits)
    return Text(f"({pct}%)", style='blue')


def quoted_text(
    _string: str,
    style: str = '',
    quote_char_style: str = 'white',
    quote_char: str = "'"
) -> Text:
    """Wrap _string in 'quote_char'. Style 'quote_char' with 'quote_char_style'."""
    quote_char_txt = Text(quote_char, style=quote_char_style)
    txt = quote_char_txt + Text(_string, style=style) + quote_char_txt
    txt.justify = 'center'
    return txt



"""
Some methods to help with the direct manipulation/processing of PyPDF's PdfObjects
"""
from dataclasses import dataclass

from pypdf.generic import ArrayObject, DictionaryObject, IndirectObject, PdfObject
from rich.text import Text

from pdfalyzer.util.adobe_strings import TYPE
from pdfalyzer.util.logging import log

BAD_NUMBER_TREE_MSG = '/Nums tree failed to be coerced to a DictionaryObject'


@dataclass
class RefAndObj:
    ref: IndirectObject
    obj: PdfObject | None


def coerce_nums_array_to_dict(nums: PdfObject | list) -> PdfObject | dict:
    """/Nums (number trees) are dict-like pairs in an array, e.g. [0, objA, 1, objB, 4, objC]."""
    bad_nums_log_msg = f"{BAD_NUMBER_TREE_MSG}:\n{nums}"

    if not (isinstance(nums, list) and len(nums) % 2 == 0):
        log.warning(bad_nums_log_msg)
        return nums

    nums_dict = {nums[i]: nums[i + 1] for i in range(0, len(nums), 2)}

    if not all(isinstance(k, int) and isinstance(v, (PdfObject, float, int, str)) for k, v in nums_dict.items()):
        log.warning(bad_nums_log_msg)
        return nums

    log.info(f"Coerced /Nums list to a dict with {len(nums_dict)} keys")
    return nums_dict


def describe_obj(_obj: PdfObject | RefAndObj | None) -> str:
    obj = _obj.obj if isinstance(_obj, RefAndObj) else _obj
    obj_str = ''

    if isinstance(obj, DictionaryObject):
        obj_type = obj.get(TYPE)
        obj_str += f"{obj_type} " if obj_type else ''

    obj_str += f"{_obj.ref.idnum} " if isinstance(_obj, RefAndObj) else ''
    obj_str += f"({type(obj).__name__}) "

    if isinstance(obj, (ArrayObject, list)):
        obj_str += f"{len(obj)} elements"

    return obj_str.strip()


def pypdf_class_name(obj: PdfObject) -> str:
    """Shortened name of type(obj), e.g. PyPDF.generic._data_structures.ArrayObject becomes Array"""
    class_pkgs = type(obj).__name__.split('.')
    class_pkgs.reverse()
    return class_pkgs[0].removesuffix('Object')



"""
Parse command line arguments for `pdfalyze` and construct the `PdfalyzerConfig` object.
"""
import sys
from argparse import ArgumentParser

from rich_argparse_plus import RichHelpFormatterPlus
from rich.prompt import Confirm
from rich.text import Text
from yaralyzer.util.argument_parser import (debug, epilog, export, parser as yaralyzer_parser,
     rules, rules, tuning, yaras)
from yaralyzer.util.constants import YARALYZER_UPPER
from yaralyzer.util.logging import log, log_console

from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.detection.constants.binary_regexes import QUOTE_PATTERNS
from pdfalyzer.util.constants import PDFALYZE
from pdfalyzer.util.output_section import ALL_STREAMS

RichHelpFormatterPlus.choose_theme('prince')

DESCRIPTION = "Explore PDF's inner data structure with absurdly large and in depth visualizations. " + \
              "Track the control flow of her darker impulses, scan rivers of her binary data for signs " + \
              "of evil sorcery, and generally peer deep into the dark heart of the Portable Document Format. " + \
              "Just make sure you also forgive her - she knows not what she does."

YARALYZER_HELP_SUFFIX = "\nThese options are only relevant when you use the --yara or --streams option."


####################################################
# Adjust Yaralyzer's option parser in a few places #
####################################################

# Add one more top level argument
yaralyzer_parser.add_argument('--password', help='only required for encrypted PDFs')

# Add one more option to the YARA rules section
yaras.add_argument('--no-default-yara-rules',
                     action='store_true',
                     help='if --yara is selected use only custom rules from --yara-file arg and not the default included YARA rules')

# Add one more option to the Debug section
debug.add_argument('--allow-missed-nodes',
                    action='store_true',
                    help='force pdfalyze to return 0 to shell even if missing nodes encountered')

# Add one more option to yaralyzer's export options
export.add_argument('-bin', '--extract-binary-streams',
                     action='store_const',
                     const='bin',
                     help='extract all binary streams in the PDF to separate files (requires pdf-parser.py)')

# Make sure --extract-binary-streams is grouped with other export options  # TODO: this really sucks.
num_args = len(yaralyzer_parser._actions)
output_dir_idx = [i for i, arg in enumerate(yaralyzer_parser._actions) if '--output-dir' in arg.option_strings][0]

yaralyzer_parser._actions = yaralyzer_parser._actions[:output_dir_idx] + \
                            [yaralyzer_parser._actions[-1]] + \
                            yaralyzer_parser._actions[output_dir_idx:-1]

assert len(yaralyzer_parser._actions) == num_args, "Number of args changed after reorder!"

# Make yara options unrequired (yaralyzer requires one of them)
rules.required = False

for action_group in yaralyzer_parser._action_groups:
    action_group._actions = yaralyzer_parser._actions

# Rename the tuning section, append info about how these are Yaralyzer options
tuning.title = f"{YARALYZER_UPPER} FINE TUNING"
tuning.description = f"{tuning.description}{YARALYZER_HELP_SUFFIX}"
rules.description = f"{rules.description}{YARALYZER_HELP_SUFFIX}"


###############################
# Pdfalyzer's argument parser #
###############################

parser = ArgumentParser(
    formatter_class=RichHelpFormatterPlus,
    description=DESCRIPTION,
    epilog=epilog(PdfalyzerConfig).rstrip(),
    parents=[yaralyzer_parser],  # NOTE: we're extending yaralyzer's parser
    add_help=False
)


# Output section selection
select = parser.add_argument_group(
    'ANALYSIS SELECTION',
    "Multiselect. Choosing nothing is choosing everything except --streams.")

select.add_argument('-d', '--docinfo', action='store_true',
                    help='show embedded document info (author, title, timestamps, etc.), streams overview, and MD5/SHA hashes')

select.add_argument('-t', '--tree', action='store_true',
                    help='show condensed tree visualization (one line per PDF object)')

select.add_argument('-r', '--rich', action='store_true',
                    help='show much larger / more detailed tree visualization (one row per PDF object property)')

select.add_argument('-f', '--fonts', action='store_true',
                    help="show info about fonts including character mappings for embedded font binaries")

select.add_argument('-y', '--yara', action='store_true',
                    help="scan the PDF with the included malicious PDF YARA rules and/or your custom YARA rules")

select.add_argument('-c', '--counts', action='store_true',
                    help='show counts of some of the properties of the objects in the PDF')

select.add_argument('-s', '--streams',
                    help="scan all the PDF's decoded/decrypted streams for sus content as well as any YARA rule matches. " +
                         "brute force is involved; output is verbose. a single OBJ_ID can be optionally provided to " +
                         "limit the output to a single internal object. try '-s -- [OTHERARGS]' if you run into an " +
                         "argument position related piccadilly.",
                    nargs='?',
                    const=ALL_STREAMS,
                    metavar='ID',
                    type=int)

select.add_argument('--extract-quoted',
                    help="extract and force decode all bytes found between this kind of quotation marks " +
                         "(requires --streams. can be specified more than once)",
                    choices=list(QUOTE_PATTERNS.keys()),
                    dest='extract_quoteds',
                    action='append')

select.add_argument('--suppress-boms', action='store_true',
                    help="don't scan streams for byte order marks (suppresses some of the --streams output)")

select.add_argument('--preview-stream-length',
                    help='number of bytes at the beginning and end of stream data to show as a preview',
                    metavar='BYTES',
                    type=int)

# Make sure the selection section is at the top
parser._action_groups = parser._action_groups[:2] + [parser._action_groups[-1]] + parser._action_groups[2:-1]
is_pdfalyze_script = parser.prog.startswith(PDFALYZE)  # startswith() bc on Windows we end up with 'pdfalyze.cmd'?



"""
Verify that the PDF tree is complete/contains all the nodes in the PDF file.
"""
from dataclasses import dataclass, field
from types import NoneType

from pypdf.generic import (ArrayObject, BooleanObject, DictionaryObject, IndirectObject, NameObject,
     NullObject, NumberObject, PdfObject, StreamObject)
from rich.markup import escape
from yaralyzer.util.logging import log

from pdfalyzer.util.adobe_strings import PREV, SIZE, TYPE, XREF, XREF_STREAM
from pdfalyzer.util.helpers.pdf_object_helper import describe_obj

OK_UNPLACED_TYPES = (BooleanObject, NameObject, NoneType, NullObject, NumberObject)


@dataclass
class PdfTreeVerifier:
    """
    Class to verify that the PDF tree is complete/contains all the nodes in the PDF file.

    Attributes:
        pdfalyzer (Pdfalyzer): The Pdfalyzer instance being verified
    """
    pdfalyzer: 'Pdfalyzer'  # noqa: F821

    def __post_init__(self):
        self._verify_unencountered_are_untraversable()

    def log_missing_node_warnings(self) -> None:
        """Log information about nodes that failed to be placed in the PDF tree."""
        print('')
        unplaced_encountered_nodes = self.pdfalyzer.unplaced_encountered_nodes()

        if len(unplaced_encountered_nodes) > 0:
            msg = f"Some nodes were traversed but never placed: {escape(str(unplaced_encountered_nodes))}\n\n" + \
                   "For link nodes like /First, /Next, /Prev, and /Last this might be no big deal - depends " + \
                   "on the PDF. But for other node typtes this could indicate missing data in the tree.\n"
            log.warning(msg)

        missing_node_ids = self.pdfalyzer.missing_node_ids()
        notable_missing_node_ids = self.notable_missing_node_ids()
        indeterminate_missing_node_ids = [id for id in missing_node_ids if id in self.pdfalyzer._indeterminate_ids]
        all_missing_nodes_msg = lambda s: f"{len(missing_node_ids)} missing node ids{s}: {missing_node_ids}"

        if notable_missing_node_ids:
            log.warning(f"Found {len(notable_missing_node_ids)} important missing node IDs: {notable_missing_node_ids}")

            if missing_node_ids != notable_missing_node_ids:
                log.warning(f"All of the {all_missing_nodes_msg(' including empty objs')}")
        elif missing_node_ids:
            log.warning(f"Identified {all_missing_nodes_msg(' but they are all scalars or empty objects')}")

        for idnum in self.pdfalyzer.missing_node_ids():
            obj = self.pdfalyzer.ref_and_obj_for_id(idnum).obj
            log.warning(f"Missing node ID {idnum} ({type(obj).__name__})")

        if (nodes_without_parents := self.pdfalyzer.nodes_without_parents()):
            node_id_to_child_count = {n.idnum: f"has {len(n.children)} children" for n in nodes_without_parents}
            log.warning(f"These node IDs were parsed but have no parent:\n{node_id_to_child_count}\n")

    def notable_missing_node_ids(self) -> list[int]:
        """Missing idnums that aren't NullObject, NumberObject, etc."""
        notable_ids = []

        for idnum in self.pdfalyzer.missing_node_ids():
            obj = self.pdfalyzer.ref_and_obj_for_id(idnum).obj
            msg = f"Missing node {idnum} {describe_obj(obj)}"

            if isinstance(obj, OK_UNPLACED_TYPES):
                log.info(f"{msg} but it's an acceptable type (value={obj})")
            elif isinstance(obj, (ArrayObject, DictionaryObject)) and len(obj) == 0:
                if isinstance(obj, StreamObject) and len(obj.get_data()) > 0:
                    notable_ids.append(idnum)
                else:
                    log.info(f"{msg} but it's empty, so it's ok")
            else:
                notable_ids.append(idnum)

        return notable_ids

    def was_successful(self):
        """Return True if no unplaced nodes or missing node IDs."""
        return (len(self.pdfalyzer.unplaced_encountered_nodes() + self.notable_missing_node_ids())) == 0

    def _verify_unencountered_are_untraversable(self) -> None:
        """
        Make sure any PDF object IDs we can't find in tree are /ObjStm or /Xref nodes and
        make a final attempt to place a few select kinds of nodes.
        """
        if self.pdfalyzer.max_generation > 0:
            log.warning(f"Verification doesn't check revisions but this PDF's generation is {self.pdfalyzer.max_generation}")  # noqa: E501

        for idnum in self.pdfalyzer.missing_node_ids():
            ref_and_obj = self.pdfalyzer.ref_and_obj_for_id(idnum)
            obj = ref_and_obj.obj

            if obj is None:
                log.error(f"Couldn't verify elementary obj with id {idnum} is properly in tree")
                continue
            elif isinstance(obj, (NumberObject, NameObject)):
                log.info(f"Obj {idnum} is a {type(obj)} w/value {obj}; if relationshipd by /Length etc. this is a nonissue but maybe worth doublechecking")  # noqa: E501
                continue
            elif not isinstance(obj, dict):
                log.error(f"Obj {idnum} ({obj}) of type {type(obj)} isn't dict, cannot determine if it should be in tree")  # noqa: E501
                continue
            elif TYPE not in obj:
                msg = f"Obj {idnum} has no {TYPE} and is not in tree. Either a loose node w/no data or an error in pdfalyzer."  # noqa: E501
                msg += f"\nHere's the contents for you to assess:\n{obj}"
                log.warning(msg)
                continue

            obj_type = obj.get(TYPE)

            if obj_type == XREF:
                placeable = XREF_STREAM in self.pdfalyzer.pdf_reader.trailer

                for k, v in self.pdfalyzer.pdf_reader.trailer.items():
                    xref_val_for_key = obj.get(k)

                    if k in [XREF_STREAM, PREV]:
                        continue
                    elif k == SIZE:
                        if xref_val_for_key is None or v != (xref_val_for_key + 1):
                            log.info(f"{XREF} has {SIZE} of {xref_val_for_key}, trailer has {SIZE} of {v}")
                            placeable = False

                        continue
                    elif k not in obj or v != obj.get(k):
                        log.info(f"Trailer has {k} -> {v} but {XREF} obj has {obj.get(k)} at that key")
                        placeable = False

                if placeable:
                    self.pdfalyzer.pdf_tree.add_child(self.pdfalyzer._build_or_find_node(ref_and_obj.ref, XREF_STREAM))
            else:
                log.warning(f"{XREF} Obj {idnum} not found in tree!")



"""
Helper methods for dealing with lists and dicts.
"""
import json

from pypdf.generic import DictionaryObject
from yaralyzer.util.logging import log


def compare_dicts(d1: DictionaryObject, d2: DictionaryObject, already_compared_keys: list[str] | None = None) -> None:
    should_call_again = already_compared_keys is None
    already_compared_keys = already_compared_keys or []

    for k, v in d1.items():
        if k in already_compared_keys or k == '/FontDescriptor':
            continue
        elif k not in d2:
            log.warning(f"'{k}' is not in dict2")
        elif v != d2[k]:
            print('')
            log.warning(f"'{k}' has different values.\n    d1['{k}']: {v}\n    d2['{k}']: {d2[k]}\n")

        already_compared_keys.append(k)

    # call with reversed args
    if should_call_again:
        compare_dicts(d2, d1, already_compared_keys)


def prefix_keys(prefix: str, _dict: dict[str, str]) -> dict[str, str]:
    """Add `prefix` to the front of all the keys in `_dict`."""
    return {f"{prefix}{k}": v for k, v in _dict.items()}


def safe_json(obj: object) -> str:
    return json.dumps(stringify_props(obj), indent=4)


def stringify_props(obj: object) -> object:
    """Make `obj` safe for JSON export."""
    if isinstance(obj, list):
        return [stringify_props(element) for element in obj]
    elif isinstance(obj, dict):
        return {str(k): stringify_props(v) for k, v in obj.items()}
    elif '__dict__' in dir(obj):
        return {str(k): stringify_props(v) for k, v in vars(obj).items()}
    elif isinstance(obj, (float, int, str)):
        return obj
    else:
        try:
            return json.dumps(obj)
        except Exception as e:
            return str(obj)


def without_falsey(_list: list) -> list:
    """Return `_list` without falsey elements (None, empty string, empty dict/list)."""
    return [e for e in _list if e]



"""
String constants specified in the Adobe specs for PDFs, fonts, etc.
"""
from pypdf.constants import (CatalogDictionary, ImageAttributes, PageAttributes,
     PagesAttributes, Resources)

from pdfalyzer.util.helpers.string_helper import is_prefixed_by_any

# Fake PDF instructions used to create more explanatory tables/trees/addresses/etc.
ARRAY_ELEMENT = '/ArrayElement'
TRAILER = '/Trailer'
UNLABELED = '/UnlabeledArrayElement'

# Actual PDF instructions
AA                   = CatalogDictionary.AA  # Automatic Action
ACRO_FORM            = CatalogDictionary.ACRO_FORM  # Can trigger Javascript on open
ANNOT                = '/Annot'
ANNOTS               = '/Annots'
CATALOG              = '/Catalog'
COLOR_SPACE          = Resources.COLOR_SPACE
CONTENTS             = '/Contents'
D                    = '/D'  # Destination, usually of a link or action
DESCENDANT_FONTS     = '/DescendantFonts'
DEST                 = '/Dest'  # Similar to /D?
ENCODING             = '/Encoding'
EXT_G_STATE          = Resources.EXT_G_STATE
FALSE                = '/False'
FIELDS               = '/Fields'
FIRST                = '/First'
FONT                 = Resources.FONT
FONT_FILE            = '/FontFile'
FONT_FILE2           = FONT_FILE + '2'
FONT_FILE3           = FONT_FILE + '3'
FONT_DESCRIPTOR      = '/FontDescriptor'
GO_TO                = '/GoTo'
GO_TO_E              = '/GoToE'  # Go to a destination in an embedded file
GO_TO_R              = '/GoToR'  # Go to a destination in another document.
GROUP                = '/Group'
IMAGE                = '/Image'
IMPORT_DATA          = '/ImportData'
INFO                 = '/Info'
JAVASCRIPT           = '/JavaScript'
JS                   = '/JS'
K                    = '/K'  # Equivalent of /Kids for /StructElem
KIDS                 = PagesAttributes.KIDS
LAST                 = '/Last'
LAUNCH               = '/Launch'
METADATA             = CatalogDictionary.METADATA
NAME                 = '/Name'
NAMED                = '/Named'
NAMES                = '/Names'
NEXT                 = '/Next'
NUMS                 = '/Nums'
OBJ                  = '/Obj'
OBJ_STM              = '/ObjStm'
# TODO: /Pg refs could be the parents of /OBJR?
OBJR                 = '/OBJR'  # Object reference to "an entire PDF object"
OPEN_ACTION          = CatalogDictionary.OPEN_ACTION
OUTLINES             = CatalogDictionary.OUTLINES
P                    = '/P'  # Equivalent of /Parent for /StructElem
PAGE                 = '/Page'
PAGES                = '/Pages'
PARENT               = PagesAttributes.PARENT
PARENT_TREE          = '/ParentTree'
PARENT_TREE_NEXT_KEY = '/ParentTreeNextKey'
PG                   = '/Pg'  # Page ref for OBJR
PREV                 = '/Prev'
RENDITION            = '/Rendition'
RESOURCES            = PageAttributes.RESOURCES
S                    = '/S'  # Equivalent of /Subtype for /StructElem
SIZE                 = '/Size'
STRUCT_ELEM          = '/StructElem'
STRUCT_PARENT        = '/StructParent'
STRUCT_TREE_ROOT     = '/StructTreeRoot'
SUBMIT_FORM          = '/SubmitForm'
SUBTYPE              = ImageAttributes.SUBTYPE
THREAD               = '/Thread'
TITLE                = '/Title'
TO_UNICODE           = '/ToUnicode'
TRUE                 = '/True'
TYPE                 = PageAttributes.TYPE
TYPE1_FONT           = '/Type1'
URI                  = '/URI'
W                    = '/W'  # Equivalen of /Widths in some situations
WIDGET               = '/Widget'
WIDTHS               = '/Widths'
XOBJECT              = Resources.XOBJECT
XREF                 = '/XRef'
XREF_STREAM          = '/XRefStm'

# There can be up to 3 /Length1, Length2, etc. keys depending on the type of font.
# They indicate points in the binary stream where different sections of the font definition
# can be found.
FONT_LENGTHS = [f'/Length{i + 1}' for i in range(3)]
FONT_FILE_KEYS = [FONT_FILE, FONT_FILE2, FONT_FILE3]

# Instructions to flag when scanning stream data for malicious content. The leading
# front slash will be removed when pattern matching.
DANGEROUS_PDF_KEYS = [
    # AA,  # AA is too generic; can't afford to remove the frontslash
    ACRO_FORM,
    GO_TO_E,
    GO_TO_R,
    IMPORT_DATA,
    JAVASCRIPT,
    JS,
    LAUNCH,
    OPEN_ACTION,
    RENDITION,
    THREAD,
    SUBMIT_FORM
]

# Adobe font instruction that begins the binary (usually encrypted) section of the font definition
CURRENTFILE_EEXEC = b'currentfile eexec'

# A node with this label is really just a non-tree link between nodes
# TODO: not sure these really need to be separate from NON_TREE_REFERENCES
LINK_NODE_KEYS = [
    D,
    DEST,
    NUMS
]

# Some references are never part of a parent/child relationship in the tree
NON_TREE_REFERENCES = [
    OPEN_ACTION,
    LAST,
    NEXT,
    PREV,
]

# Some PdfObjects can't be properly placed in the tree until the entire tree is parsed
INDETERMINATE_REF_KEYS = [
    ANNOTS,     # At least when it appears in a page
    COLOR_SPACE,
    D,
    DEST,
    EXT_G_STATE,
    FIELDS,     # At least for  /AcroForm
    FIRST,
    FONT,
    NAMES,
    OPEN_ACTION,
    P,          # At least for widgets...
    RESOURCES,
    XOBJECT,
    UNLABELED,  # TODO: this might be wrong? maybe this is where the /Resources actually live?
]

INDETERMINATE_PREFIXES = [p for p in INDETERMINATE_REF_KEYS if len(p) > 2]
NON_TREE_KEYS = LINK_NODE_KEYS + NON_TREE_REFERENCES
PAGE_AND_PAGES = [PAGE, PAGES]

MULTI_REF_NODE_TYPES = [
    NUMS,
    PARENT_TREE
]

# Address reference keys that don't always appear (example: sometimes there is only a link to FIRST and LAST)
# and none of the NEXT/PREV nodes between FIRST and LAST.
IMPERMANENT_KEYS = [
    FIRST,
    LAST,
    NEXT,
    PREV,
]

# Address reference keys that adon't always appear or b) can appear more than once pointing at same node
NON_STANDARD_ADDRESS_NODES = IMPERMANENT_KEYS + MULTI_REF_NODE_TYPES

# Keys that sometimes create a lot of log noise with multiple refs
MULT_REF_RESOURCE_KEYS = [COLOR_SPACE, '/Shading']


def has_indeterminate_prefix(address: str) -> bool:
    return is_prefixed_by_any(address, INDETERMINATE_PREFIXES)



"""
Various text formatting/styling/manipulating methods.
"""
import re
from pprint import PrettyPrinter
from typing import Pattern

from pypdf.generic import IndirectObject

from yaralyzer.output.console import console_width
from yaralyzer.util.helpers.string_helper import INDENT_DEPTH, INDENT_SPACES, INDENTED_JOINER, NON_WORD_CHAR_REGEX, indented

ARRAY_IDX_REGEX = re.compile(r"\[\d+\]")
PRETTY_PRINT_WIDTH = 60
DIGIT_REGEX = re.compile(r"\d+")

# Pretty Printer
pp = PrettyPrinter(
    indent=INDENT_DEPTH,
    width=PRETTY_PRINT_WIDTH,
    sort_dicts=True
)


def all_strings_are_same_ignoring_numbers(strings: list[str]) -> bool:
    """Returns true if string addresses are same except for digits."""
    return len(set([replace_digits(s) for s in strings])) == 1


def bracketed(index: int | str) -> str:
    """Surround index with [ and ]."""
    return f"[{index}]"


def class_name_regex(t: type) -> re.Pattern:
    return re.compile(t.__name__)


def coerce_address(address: str | int) -> str:
    return bracketed(address) if isinstance(address, int) else address


def count_pattern_matches_in_text(pattern: str, text: str) -> int:
    return count_regex_matches_in_text(re.compile(pattern), text)


def count_regex_matches_in_text(regex: Pattern, text: str) -> int:
    """For use when you precompile the regex."""
    return sum(1 for _ in regex.finditer(text))


def exception_str(e: Exception) -> str:
    """A string with the type and message."""
    return f"{type(e).__name__}: {e}"


def generate_hyphen_line(width: int | None = None, title: str | None = None):
    """e.g. '-----------------BEGIN-----------------'"""
    width = width or console_width()

    if title is None:
        return '-' * width

    side_hyphens = int((width - len(title)) / 2) * '-'
    line = side_hyphens + title + side_hyphens
    return line if len(line) == width else line + '-'


def has_a_common_substring(strings: list[str]) -> bool:
    return all([is_substring_of_longer_strings_in_list(s, strings) for s in strings])


def highlight_pattern(regex: re.Pattern | str) -> str:
    """Build a rich.Highlighter style pattern, e.g. `(?P<stream_object>((De|En)coded)?StreamObject)`."""
    pattern = regex.pattern if isinstance(regex, re.Pattern) else regex
    label = regex_to_capture_group_label(regex)

    if len(pattern) <= 2:
        pattern = fr"^{pattern}$"
    else:
        pattern = fr"{pattern}\b".removeprefix('/')

        if not pattern.startswith('^'):
            pattern = fr"[\b/]{pattern}"

    return fr"(?P<{label}>{pattern})"


def is_array_idx(address: str) -> bool:
    """True if address looks like '[23]'."""
    return bool(ARRAY_IDX_REGEX.match(address))


def is_prefixed_by_any(_string: str, prefixes: list[str]) -> bool:
    """Returns True if _string starts with anything in 'prefixes'."""
    return any([_string.startswith(prefix) for prefix in prefixes])


def is_substring_of_longer_strings_in_list(_string: str, strings: list[str]) -> bool:
    """Return True if '_string' is a substring of all the 'strings' longer than '_string'."""
    longer_strings = [s for s in strings if len(s) > len(_string)]
    return all([_string in longer_string for longer_string in longer_strings])


def numbered_list(objs: list, indent: int = 4) -> str:
    list_str = '\n'.join([f"[{i + 1}] {e}" for i, e in enumerate(objs)])
    return indented(list_str, spaces=indent)


def props_string(obj: object, keys: list[str] | None = None, joiner: str = ', ') -> str:
    prefix = INDENT_SPACES if '\n' in joiner else ''
    return prefix + joiner.join(props_strings(obj, keys))


def props_string_indented(obj: object, keys: list[str] | None = None) -> str:
    return props_string(obj, keys, INDENTED_JOINER)


def props_strings(obj: object, keys: list[str] | None = None) -> list[str]:
    """Get props of 'obj' in the format ["prop1=5", "prop2='string'"] etc."""
    props = []

    for k in (keys or [k for k in vars(obj).keys()]):
        value = getattr(obj, k)
        value = f"'{value}'" if isinstance(value, str) else value
        value = repr(value) if isinstance(value, IndirectObject) else value
        props.append(f"{k}={value}")

    return props


def regex_to_capture_group_label(pattern: re.Pattern | str) -> str:
    pattern = pattern.pattern if isinstance(pattern, re.Pattern) else pattern
    return NON_WORD_CHAR_REGEX.sub('', pattern.replace('|', '_'))


def replace_digits(string_with_digits: str) -> str:
    """Turn all digits to X chars in a string."""
    return DIGIT_REGEX.sub('x', string_with_digits)


def root_address(_string: str) -> str:
    """Strip the bracketed part off an address, e.g. '/Root[1]' => '/Root'."""
    return _string.split('[')[0]



"""
Regex patterns and a LogHighlighter that work with the rich.Highlighter approach of coloring
text output.
"""
import re

from rich.markup import escape
from rich.highlighter import ReprHighlighter
from yaralyzer.util.logging import log_console

from pdfalyzer.util.helpers.collections_helper import prefix_keys
from pdfalyzer.util.helpers.rich_helper import vertically_padded_panel

PDF_OBJ_STYLE_PREFIX = 'pdf.'
PYPDF_LOG_PFX_PATTERN = r"\(pypdf\)"

# Styles
CHILD_STYLE = 'orange3 bold'
INDIRECT_OBJ_STYLE = 'light_coral'  # Formerly 'color(225)'
PARENT_STYLE = 'violet'
PDF_ARRAY_STYLE = 'color(143)'  # color(120)
PDF_DICTIONARY_STYLE = 'color(64)'

# Copied from https://rich.readthedocs.io/en/latest/_modules/rich/highlighter.html#Highlighter
# so we can get rid of a couple of the patterns we don't want.
DEFAULT_REPR_HIGHLIGHTER_PATTERNS = [
    r"(?P<tag_start><)(?P<tag_name>[-\w.:|]*)(?P<tag_contents>[\w\W]*)(?P<tag_end>>)",
    r'(?P<attrib_name>[\w_]{1,50})=(?P<attrib_value>"?[\w_]+"?)?',
    r"(?P<brace>[][{}()])",
    r"(?P<ipv4>[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})",
    r"(?P<ipv6>([A-Fa-f0-9]{1,4}::?){1,7}[A-Fa-f0-9]{1,4})",
    r"(?P<eui64>(?:[0-9A-Fa-f]{1,2}-){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){3}[0-9A-Fa-f]{4})",
    r"(?P<eui48>(?:[0-9A-Fa-f]{1,2}-){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){2}[0-9A-Fa-f]{4})",
    r"(?P<uuid>[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})",
    r"(?P<call>[\w.]*?)\(",
    r"\b(?P<bool_true>True)\b|\b(?P<bool_false>False)\b|\b(?P<none>None)\b",
    r"(?P<ellipsis>\.\.\.)",
    r"(?P<number_complex>(?<!\w)(?:\-?[0-9]+\.?[0-9]*(?:e[-+]?\d+?)?)(?:[-+](?:[0-9]+\.?[0-9]*(?:e[-+]?\d+)?))?j)",
    r"(?P<number>(?<!\w)\-?[0-9]+\.?[0-9]*(e[-+]?\d+?)?\b|0x[0-9a-fA-F]*)",
    # r"(?P<path>\B(/[-\w._+]+)*\/)(?P<filename>[-\w._+]*)?",
    r"(?<![\\\w])(?P<str>b?'''.*?(?<!\\)'''|b?'.*?(?<!\\)'|b?\"\"\".*?(?<!\\)\"\"\"|b?\".*?(?<!\\)\")",
    r"(?P<url>(file|https|http|ws|wss)://[-0-9a-zA-Z$_+!`(),.?/;:&=%#~@]*)",
]

# Our custom log highlight patterns
LOG_HIGHLIGHT_PATTERNS = DEFAULT_REPR_HIGHLIGHTER_PATTERNS + [
    r"(?P<child>[cC]hild(ren)?)",
    r"(?P<indeterminate>[Ii]ndeterminate( ?[nN]odes?)?)",
    r"(?P<parent>[pP]arents?)",
    fr"(?P<pypdf_line>{PYPDF_LOG_PFX_PATTERN} .*)",
    fr"(?P<pypdf_prefix>{PYPDF_LOG_PFX_PATTERN})",
    r"(?P<relationship>Relationship( of)?)",
    r"(?P<relationship>via symlink|parent/child|child/parent)",
]

# Logger highlights
LOG_HIGHLIGHT_STYLES = {
    "child": CHILD_STYLE,
    "indeterminate": 'bright_black',
    "parent": PARENT_STYLE,
    "pypdf_line": "dim",
    "pypdf_prefix": "light_slate_gray",
    "relationship": 'light_pink4',
    # Overload default ReprHighlighter theme elements
    'call': 'magenta',
    'ipv4': 'cyan',
    'ipv6': 'cyan',
}


# Augment the standard ReprHighlighter
class LogHighlighter(ReprHighlighter):
    highlights: list[re.Pattern]

    @classmethod
    def set_highlights(cls, patterns: list[str]) -> None:
        """Compile strings to regexes."""
        cls.highlights = [re.compile(p) for p in (patterns)]

    @classmethod
    def get_style(cls, for_str: str) -> str:
        """Return the first style that matches the 'for_str'."""
        for highlight in cls.highlights:
            if (match := highlight.search(for_str)):
                return cls.base_style + next(k for k in match.groupdict().keys())

        return ''

    @classmethod
    def prefix_styles(cls, styles: dict[str, str]) -> dict[str, str]:
        """Prepend this highlighter's `base_style` to all keys in the `styles` dict."""
        return prefix_keys(cls.base_style, styles)

    @classmethod
    def prefixed_style(cls, style: str) -> str:
        """Prepend this highlighter's `base_style` to `style` string."""
        return cls.base_style + style

    @classmethod
    def _debug_highlight_patterns(cls):
        log_console.print(vertically_padded_panel(f"{cls.__name__}.highlights Patterns (base: '{cls.base_style}')"))

        for pattern in cls.highlights:
            log_console.print(f"   - '{escape(str(pattern))}'")


class PdfHighlighter(LogHighlighter):
    base_style = PDF_OBJ_STYLE_PREFIX

    @classmethod
    def prefixed_style(cls, style: str) -> str:
        """Prepend this highlighter's `base_style` to `style` string, removing first slash."""
        return cls.base_style + style.removeprefix('/')


# Instantiate highlighters
log_highlighter = LogHighlighter()
pdf_highlighter = PdfHighlighter()



"""
Methods to create the rich table view for a PdfTreeNode.
"""
from rich.text import Text
from yaralyzer.output.theme import BYTES_NO_DIM
from yaralyzer.util.helpers.bytes_helper import NEWLINE_BYTE, clean_byte_string, hex_text
from yaralyzer.util.helpers.rich_helper import size_text
from yaralyzer.util.logging import log

HEX = 'Hex'
STREAM = 'Stream'
STREAM_PREVIEW_LENGTH_IN_TABLE = 500
PREVIEW_STYLES = {HEX: BYTES_NO_DIM, STREAM: 'bytes'}


# TODO: this should be a method on the Objstm or other StreamNode extension to PdfTreeNode or PdfObjProps
# branch: pypdf_6.6.0__local_pypdf_changes__objstm
def get_stream_preview_rows(node: 'PdfTreeNode') -> list[list[Text]]:  # noqa: F821
    """Get rows that preview the stream data"""
    if node.stream_length == 0:
        return []
    elif node.stream_data is None or len(node.stream_data) == 0:
        log.warning(node.__rich__().append(' is a stream object but had no stream data'))
        return []

    stream_preview = node.stream_data[:STREAM_PREVIEW_LENGTH_IN_TABLE]
    stream_preview_length = len(stream_preview)
    return_rows: list[list[Text]] = []

    if isinstance(node.stream_data, bytes):
        stream_preview_hex = hex_text(stream_preview).plain
        stream_preview_lines = [clean_byte_string(line) for line in stream_preview.split(NEWLINE_BYTE)]
        stream_preview_string = "\n".join(stream_preview_lines)
    else:
        stream_preview_hex = f"N/A (Stream data is type '{type(node.stream_data).__name__}', not bytes)"
        stream_preview_string = stream_preview

    def add_preview_row(hex_or_stream: str, stream_string: str):
        if stream_preview_length < STREAM_PREVIEW_LENGTH_IN_TABLE:
            row_label = "Data" if hex_or_stream != HEX else ' View'
        else:
            row_label = "Preview" if hex_or_stream != HEX else ' Preview'
            stream_string += '...'

        style = PREVIEW_STYLES[hex_or_stream]
        row_label = f"{hex_or_stream}{row_label}\n  ({stream_preview_length} bytes)"
        return_rows.append([Text(row_label, 'grey'), Text(stream_string, style)])

    add_preview_row(STREAM, stream_preview_string)
    add_preview_row(HEX, stream_preview_hex)
    return_rows.append([Text('StreamLength', style='grey'), size_text(len(node.stream_data))])
    return return_rows



"""
Color theme for pdfalyzer including functions to help decide on node colors etc.

Rich colors: https://rich.readthedocs.io/en/stable/appendix/colors.html
TODO: interesting colors # row_styles[0] = 'reverse bold on color(144)' <-
"""
import re
from collections import namedtuple
from numbers import Number
from types import NoneType
from typing import Any

from pypdf.generic import (ArrayObject, ByteStringObject, EncodedStreamObject, IndirectObject, NullObject,
     StreamObject, TextStringObject)
from rich.theme import Theme
from yaralyzer.output.console import console
from yaralyzer.output.theme import BYTES_NO_DIM, YARALYZER_THEME_DICT
from yaralyzer.util.logging import log_console

from pdfalyzer.output.highlighter import (CHILD_STYLE, INDIRECT_OBJ_STYLE, LOG_HIGHLIGHT_PATTERNS,
     LOG_HIGHLIGHT_STYLES, PARENT_STYLE, PDF_ARRAY_STYLE, PDF_DICTIONARY_STYLE, LogHighlighter, PdfHighlighter)
from pdfalyzer.util import adobe_strings
from pdfalyzer.util.helpers.collections_helper import safe_json
from pdfalyzer.util.helpers.rich_helper import vertically_padded_panel
from pdfalyzer.util.helpers.string_helper import highlight_pattern

ClassStyle = namedtuple('ClassStyle', ['cls', 'style'])

# For highlighting / naming styles
PDF_OBJ_STYLE_PFX = 'pdfobj.'

# Colors / PDF object styles
DEFAULT_LABEL_STYLE = 'yellow'
DEFAULT_OBJ_TYPE_STYLE = 'bright_yellow'
FONT_FILE_BLUE = 'steel_blue1'
FONT_OBJ_BLUE = 'deep_sky_blue4 bold'
INFO_OBJ_STYLE = 'yellow4'
LINK_OBJ_STYLE = 'grey46'
METADATA_STYLE = 'color(35)'
NULL_STYLE = 'grey23'
PAGE_OBJ_STYLE = 'light_salmon3'
PDF_NON_TREE_REF_STYLE = 'color(243)'  # grey46?
RED_ALERT_BASE_STYLE = 'blink bold red'
TRAILER_OBJ_STYLE = 'chartreuse2'
TYPE_KEY_STYLE = 'honeydew2'

PDFALYZER_THEME_DICT = {
    **YARALYZER_THEME_DICT,
    'bytes.BOM': 'bright_green',
    # charmap
    'charmap.title': 'color(25)',
    'charmap.prepared_title': 'color(23)',
    'charmap.prepared': 'color(106) dim',
    'charmap.byte': 'color(58)',
    'charmap.char': 'color(120) bold',
    # fonts
    'font.property': 'color(135)',
    'font.title': 'reverse dark_blue on color(253)',
    # design elements
    'subtable': 'color(8) on color(232)',
    # warn log events
    'warn': 'bright_yellow',
    # error log events
    'red_alert': f'{RED_ALERT_BASE_STYLE} reverse on white',
}

PDF_OBJ_TYPE_STYLES = [
    ClassStyle(IndirectObject, INDIRECT_OBJ_STYLE),
    ClassStyle(ByteStringObject, YARALYZER_THEME_DICT['bytes.title']),
    ClassStyle(EncodedStreamObject, YARALYZER_THEME_DICT['bytes']),
    ClassStyle(StreamObject, BYTES_NO_DIM),
    ClassStyle(TextStringObject, YARALYZER_THEME_DICT['grey.light']),
    ClassStyle(ArrayObject, PDF_ARRAY_STYLE),
    ClassStyle(NullObject, NULL_STYLE),
    ClassStyle(NoneType, NULL_STYLE),
]

PDF_OBJ_TYPE_STYLE_DICT = {f"{cs.cls.__name__}": cs.style for cs in PDF_OBJ_TYPE_STYLES}

# Subclasses of the key type will be styled with the value string
OBJ_TYPE_STYLES = PDF_OBJ_TYPE_STYLES + [
    ClassStyle(Number, 'cyan bold'),
    ClassStyle(dict, PDF_DICTIONARY_STYLE),
    ClassStyle(list, PDF_ARRAY_STYLE),
    ClassStyle(tuple, PDF_ARRAY_STYLE),
    ClassStyle(str, 'bright_white bold'),
]

# Add styles for all NON_TREE_REFERENCES first because /OpenAction is one such action
NODE_STYLES_BASE_DICT = {
    key: PDF_NON_TREE_REF_STYLE
    for key in (adobe_strings.NON_TREE_REFERENCES + adobe_strings.LINK_NODE_KEYS + [adobe_strings.FIRST])
}

# Order matters - first match will be the style
NODE_STYLES_BASE_DICT.update({
    adobe_strings.AA:                                          RED_ALERT_BASE_STYLE,
    adobe_strings.JAVASCRIPT:                                  RED_ALERT_BASE_STYLE,
    adobe_strings.JS:                                          RED_ALERT_BASE_STYLE,
    adobe_strings.OPEN_ACTION:                                 RED_ALERT_BASE_STYLE,
    adobe_strings.GO_TO_R:                                     RED_ALERT_BASE_STYLE,
    '/Action':                                                 'red',
    # Fonts
    adobe_strings.FONT_DESCRIPTOR:                             'cornflower_blue',
    f'{adobe_strings.FONT_FILE}':                              FONT_FILE_BLUE,
    f'{adobe_strings.FONT_FILE}2':                             FONT_FILE_BLUE,
    f'{adobe_strings.FONT_FILE}3':                             FONT_FILE_BLUE,
    '/FontName':                                               FONT_OBJ_BLUE,
    adobe_strings.FONT:                                        FONT_OBJ_BLUE,  # After other /Font styles so it matches after
    '/BaseFont':                                               FONT_OBJ_BLUE,
    adobe_strings.DESCENDANT_FONTS:                            f"{FONT_OBJ_BLUE} dim",
    '/CharProc':                                               'dark_cyan',
    adobe_strings.TO_UNICODE:                                  'grey30',
    adobe_strings.ENCODING:                                    YARALYZER_THEME_DICT['encoding.header'],
    adobe_strings.WIDTHS:                                      'color(67)',
    adobe_strings.W:                                           'color(67)',
    adobe_strings.RESOURCES:                                   'magenta',
    # Top level nodes
    '/Root':                                                   TRAILER_OBJ_STYLE,
    adobe_strings.CATALOG:                                     'color(47)',
    adobe_strings.CONTENTS:                                    'medium_purple1',
    adobe_strings.TRAILER:                                     TRAILER_OBJ_STYLE,
    adobe_strings.INFO:                                        INFO_OBJ_STYLE,
    adobe_strings.OUTLINES:                                    INFO_OBJ_STYLE,
    adobe_strings.METADATA:                                    METADATA_STYLE,
    '/ViewerPreferences':                                      METADATA_STYLE,
    adobe_strings.OBJ_STM:                                     YARALYZER_THEME_DICT['bytes'],
    # Data nodes
    adobe_strings.ANNOTS:                                      'deep_sky_blue4',
    adobe_strings.ANNOT:                                       'color(24)',
    adobe_strings.NAMES:                                       'white',
    adobe_strings.NUMS:                                        'grey23',
    adobe_strings.UNLABELED:                                   'grey35 reverse',
    adobe_strings.XOBJECT:                                     'grey37',
    adobe_strings.XREF:                                        'color(148)',
    # Images
    adobe_strings.IMAGE:                                       'medium_violet_red',
    adobe_strings.URI:                                         'color(244)',
    # Pages
    adobe_strings.PAGES:                                       'dark_orange3',
    adobe_strings.PAGE:                                        PAGE_OBJ_STYLE,
    adobe_strings.PG:                                          PAGE_OBJ_STYLE,
    adobe_strings.COLOR_SPACE:                                 'medium_orchid1',
    # Parents
    adobe_strings.KIDS:                                        CHILD_STYLE,
    adobe_strings.PARENT:                                      PARENT_STYLE,
    adobe_strings.PARENT_TREE:                                 PARENT_STYLE,
    adobe_strings.PARENT_TREE_NEXT_KEY:                        PARENT_STYLE,
    adobe_strings.STRUCT_PARENT:                               PARENT_STYLE,
    # Booleans
    adobe_strings.FALSE:                                       'bright_red',
    adobe_strings.TRUE:                                        'green bold',
    # /Type fields
    adobe_strings.SUBTYPE:                                     TYPE_KEY_STYLE,
    adobe_strings.TYPE:                                        TYPE_KEY_STYLE,
})

# Compile regexes as keys
NODE_STYLE_REGEXES = {re.compile(k): v for k, v in NODE_STYLES_BASE_DICT.items()}

# Unite class styles for things like ArrayObject with node styles for things like /Parent
NODE_STYLES_THEME_DICT = {
    **PdfHighlighter.prefix_styles({k.removeprefix('/'): v for k, v in NODE_STYLES_BASE_DICT.items()}),
    **PdfHighlighter.prefix_styles(PDF_OBJ_TYPE_STYLE_DICT)
}

# Merge all the theme dicts
LOG_THEME_DICT = LogHighlighter.prefix_styles(LOG_HIGHLIGHT_STYLES)
COMPLETE_THEME_DICT = {**PDFALYZER_THEME_DICT, **LOG_THEME_DICT, **NODE_STYLES_THEME_DICT}


# Add patterns to highlighters
LogHighlighter.set_highlights(LOG_HIGHLIGHT_PATTERNS + [highlight_pattern(k) for k in PDF_OBJ_TYPE_STYLE_DICT.keys()])
PdfHighlighter.set_highlights([highlight_pattern(r) for r in NODE_STYLE_REGEXES.keys()])

# Push themes into the console objects that manage stdout.
console.push_theme(Theme(COMPLETE_THEME_DICT))
log_console.push_theme(Theme(COMPLETE_THEME_DICT))


def get_class_style(obj: Any) -> str:
    """Style for various types of data (e.g. DictionaryObject)"""
    if obj is True:
        cls_style = 'bright_green bold'
    elif obj is False:
        cls_style = 'bright_red bold'
    else:
        cls_style = next((cs.style for cs in OBJ_TYPE_STYLES if isinstance(obj, cs.cls)), DEFAULT_OBJ_TYPE_STYLE)

    return cls_style


def get_class_style_dim(obj: Any) -> str:
    """Dim version of get_class_style() for non primitives, white for primitives"""
    if isinstance(obj, str):
        return 'color(244)'
    elif isinstance(obj, Number):
        cls_style = get_class_style(obj).replace('bold', '').strip()
    else:
        cls_style = get_class_style(obj)

    return f"{cls_style} dim"


def get_class_style_italic(obj: Any) -> str:
    return f"{get_class_style(obj)} italic"


def get_label_style(label: str) -> str:
    """Lookup a style based on the node's label string (either its type or first address)."""
    return next((v for k, v in NODE_STYLE_REGEXES.items() if k.match(label)), DEFAULT_LABEL_STYLE)


def theme_json() -> str:
    theme_dicts = {
        'COMPLETE_THEME_DICT': COMPLETE_THEME_DICT,
        'LOG_THEME_DICT': LOG_THEME_DICT,
        'NODE_STYLES_THEME_DICT': NODE_STYLES_THEME_DICT,
        'PDFALYZER_THEME_DICT': PDFALYZER_THEME_DICT,
    }

    return safe_json(theme_dicts)


def _debug_themes() -> None:
    LogHighlighter._debug_highlight_patterns()
    PdfHighlighter._debug_highlight_patterns()
    log_console.print(vertically_padded_panel('All Theme Dicts'))
    log_console.print_json(theme_json(), indent=4, sort_keys=True)



"""
Handles formatting of console text output for Pdfalyzer class.
"""
from collections import defaultdict, namedtuple
from dataclasses import dataclass, field
from typing import Callable

import yara
from anytree import LevelOrderIter, RenderTree, SymlinkNode
from anytree.render import DoubleStyle
from rich.markup import escape
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.tree import Tree
from yaralyzer.config import YaralyzerConfig
from yaralyzer.output.console import console
from yaralyzer.output.file_hashes_table import bytes_hashes_table
from yaralyzer.output.theme import BYTES_HIGHLIGHT
from yaralyzer.util.exceptions import print_fatal_error
from yaralyzer.util.helpers.rich_helper import DEFAULT_TABLE_OPTIONS, size_in_bytes_text
from yaralyzer.yara.error import yara_error_msg
from yaralyzer.yaralyzer import Yaralyzer

from pdfalyzer.binary.binary_scanner import BinaryScanner
from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.decorators.pdf_tree_node import DECODE_FAILURE_LEN, PdfTreeNode
from pdfalyzer.detection.yaralyzer_helper import get_bytes_yaralyzer, get_file_yaralyzer
from pdfalyzer.output.layout import (print_fatal_error_panel, print_section_header, print_section_subheader,
     print_section_sub_subheader)
from pdfalyzer.output.tables.decoding_stats_table import build_decoding_stats_table
from pdfalyzer.output.theme import get_label_style
from pdfalyzer.pdfalyzer import Pdfalyzer
from pdfalyzer.util.adobe_strings import DANGEROUS_PDF_KEYS, FALSE, TRUE
from pdfalyzer.util.helpers.collections_helper import safe_json
from pdfalyzer.util.helpers.string_helper import root_address
from pdfalyzer.util.logging import highlight, log

SymlinkRepresentation = namedtuple('SymlinkRepresentation', ['text', 'style'])

COUNT_LABEL_STYLE = 'navajo_white3'
HIGHLIGHT_IF = ['http', FALSE, TRUE]

count_label = lambda count_type: f"{count_type} count"
count_txt = lambda count_type: Text(count_label(count_type), style=COUNT_LABEL_STYLE)
failed_count_msg = lambda count_type: f"Failed to get {count_label(count_type)}!"


@dataclass
class PdfalyzerPresenter:
    """
    Handles formatting of console text output for Pdfalyzer class.

    Attributes:
        pdfalyzer (Pdfalyzer): Pdfalyzer for a given PDF file
        yaralyzer (Yaralyzer): Yaralyzer for a given PDF file
    """
    pdfalyzer: Pdfalyzer
    yaralyzer: Yaralyzer = field(init=False)

    def __post_init__(self):
        self.yaralyzer = get_file_yaralyzer(self.pdfalyzer.pdf_path)

    def print_everything(self) -> None:
        """Print every kind of analysis on offer to Rich console."""
        self.print_document_info()
        self.print_summary()
        self.print_tree()
        self.print_rich_table_tree()
        self.print_font_info()
        self.print_yara_results()
        self.print_non_tree_relationships()

    def print_document_info(self) -> None:
        """Print the embedded document info (author, timestamps, version, etc)."""
        print_section_header(f'Document Info for {self.pdfalyzer.pdf_basename}')
        console.print(self._metadata_table())
        console.line()
        console.print(bytes_hashes_table(self.pdfalyzer.pdf_bytes, self.pdfalyzer.pdf_basename))
        console.line()
        console.print(self._stream_objects_table())
        console.line()

    def print_tree(self) -> None:
        """Print the simple view of the PDF tree."""
        print_section_header(f'Simple tree view of {self.pdfalyzer.pdf_basename}')

        for pre, _fill, node in RenderTree(self.pdfalyzer.pdf_tree, style=DoubleStyle):
            if isinstance(node, SymlinkNode):
                symlink_rep = self._get_symlink_representation(node.parent, node)
                console.print(pre + f"[{symlink_rep.style}]{symlink_rep.text}[/{symlink_rep.style}]")
            else:
                console.print(Text(pre) + node.__rich__())

        console.print("\n\n")

    def print_rich_table_tree(self) -> None:
        """Print the rich view of the PDF tree."""
        print_section_header(f'Rich tree view of {self.pdfalyzer.pdf_basename}')
        console.print(self._generate_rich_tree(self.pdfalyzer.pdf_tree))

    def print_summary(self) -> None:
        """Print node type counts and so on."""
        print_section_header(f'PDF Node Summary for {self.pdfalyzer.pdf_basename}')
        console.print_json(safe_json(self._analyze_tree()), sort_keys=True)

    def print_font_info(self, font_idnum=None) -> None:
        """Print informatin about all fonts that appear in this PDF."""
        print_section_header(f'{len(self.pdfalyzer.font_infos)} fonts found in {self.pdfalyzer.pdf_basename}')

        if self.pdfalyzer.font_info_extraction_error:
            print_fatal_error(f"Failed to extract font information (error: {self.pdfalyzer.font_info_extraction_error})")

        for font_info in [fi for fi in self.pdfalyzer.font_infos if font_idnum is None or font_idnum == fi.idnum]:
            font_info.print_summary()

    def print_streams_analysis(self, idnum: int | None = None) -> None:
        """
        For each binary stream,
          1. Scan decompressed binary with YARA rules we applied to whole PDF (the ones in pdfalyzer/yara_rules/)
          2. Check for (and force decode) dangerous PDF instructions like /JavaScript and /OpenAction
          3. Check for (and force decode) any BOMs (byte order marks)
          4. Check for (and force decode) any sequences of bytes between quotes
        """
        print_section_header(f'Binary Stream Analysis / Extraction')
        console.print(self._stream_objects_table())

        for node in [n for n in self.pdfalyzer.stream_nodes() if idnum is None or idnum == n.idnum]:
            node_stream_bytes = node.stream_data

            if node_stream_bytes is None or node.stream_length == 0:
                print_section_sub_subheader(f"{node} stream has length 0", style='dim')
                continue

            if not isinstance(node_stream_bytes, bytes):
                log.warning(f"Stream in {node} is not bytes, it's {type(node.stream_data).__name__}. Will " \
                             "reencode for YARA but they may not be the same bytes as the original stream!")
                node_stream_bytes = node_stream_bytes.encode()

            console.line()
            print_section_subheader(f"{escape(str(node))} Summary and Analysis", style=f"{BYTES_HIGHLIGHT} reverse")
            binary_scanner = BinaryScanner(node_stream_bytes, node)
            console.print(bytes_hashes_table(binary_scanner.bytes))
            binary_scanner.print_stream_preview()
            binary_scanner.check_for_dangerous_instructions()

            if not PdfalyzerConfig.args.suppress_boms:
                binary_scanner.check_for_boms()

            if not YaralyzerConfig.args.suppress_decodes_table:
                binary_scanner.force_decode_quoted_bytes()
                console.line(2)
                console.print(build_decoding_stats_table(binary_scanner), justify='center')

    def print_yara_results(self) -> None:
        """Scan the main PDF and each individual binary stream in it with yara_rules/*.yara files."""
        try:
            print_section_header(f"YARA Scan of PDF rules for '{self.pdfalyzer.pdf_basename}'")
            YaralyzerConfig.args._yaralyzer_standalone_mode = True  # TODO: using 'standalone mode' like this kind of sucks
            self.yaralyzer.yaralyze()
        except yara.Error as e:
            console.print_exception()
            print_fatal_error_panel(yara_error_msg(e))
            return

        YaralyzerConfig.args._yaralyzer_standalone_mode = False
        console.line(2)

        for node in self.pdfalyzer.stream_nodes():
            if node.stream_length == DECODE_FAILURE_LEN:
                log.warning(f"{node} binary stream could not be extracted")
            elif node.stream_length == 0 or node.stream_data is None:
                log.debug(f"No binary to scan for {node}")
            else:
                get_bytes_yaralyzer(node.stream_data, str(node)).yaralyze()
                console.line(2)

    def print_non_tree_relationships(self) -> None:
        """Print the inter-node, non-tree relationships for all nodes in the tree. Debugging method."""
        print_section_header(f"Non-tree Relationships for '{self.pdfalyzer.pdf_basename}'")

        for node in LevelOrderIter(self.pdfalyzer.pdf_tree):
            if len(node.non_tree_relationships) == 0:
                continue

            console.line(2)
            console.print(Panel(f"Non tree relationships for {node}", expand=False, **DEFAULT_TABLE_OPTIONS))
            node.print_non_tree_relationships()

    def _analyze_tree(self) -> dict:
        """Generate a dict with some basic data points about the PDF tree"""
        keys_encountered = defaultdict(int)
        node_count = 0
        node_labels = defaultdict(int)
        node_types = defaultdict(int)
        pdf_object_types = defaultdict(int)

        for node in self.pdfalyzer.node_iterator():
            node_count += 1
            node_labels[node.label] += 1
            node_types[node.type] += 1
            pdf_object_types[type(node.obj).__name__] += 1

            if isinstance(node.obj, dict):
                for k in node.obj.keys():
                    keys_encountered[k] += 1

        return {
            'keys_encountered': keys_encountered,
            'node_count': node_count,
            'node_labels': node_labels,
            'node_types': node_types,
            'pdf_object_types': pdf_object_types,
        }

    def _generate_rich_tree(self, node: PdfTreeNode, tree: Tree | None = None) -> Tree:
        """Recursively generates a rich.tree.Tree object from 'node' and its children."""
        tree = tree or Tree(node.as_tree_node_table(self.pdfalyzer))

        for child in node.children:
            if isinstance(child, SymlinkNode):
                symlink_rep = self._get_symlink_representation(node, child)
                tree.add(Panel(symlink_rep.text, style=symlink_rep.style, expand=False, **DEFAULT_TABLE_OPTIONS))
                continue

            child_branch = tree.add(child.as_tree_node_table(self.pdfalyzer))
            self._generate_rich_tree(child, child_branch)

        return tree

    def _get_symlink_representation(self, from_node: PdfTreeNode, to_node: SymlinkNode) -> SymlinkRepresentation:
        """Returns a tuple (symlink_text, style) that can be used for pretty printing, tree creation, etc"""
        reference_key = str(to_node.address_of_this_node_in_other(from_node))
        pdf_instruction = root_address(reference_key)  # In case we ended up with a [0] or similar

        if pdf_instruction in DANGEROUS_PDF_KEYS:
            symlink_style = 'red_alert'
        else:
            symlink_style = get_label_style(to_node.label) + ' dim'

        symlink_str = f"{escape(reference_key)} [bright_white]=>[/bright_white]"
        symlink_str += f" {escape(str(to_node.target))} [grey](Non Child Reference)[/grey]"
        return SymlinkRepresentation(symlink_str, symlink_style)

    def _stream_objects_table(self) -> Table:
        """Build a table of stream objects and their lengths."""
        table = Table(
            'Stream Length',
            'Node',
            title=' Embedded Streams',
            title_style='grey',
            title_justify='left',
            **DEFAULT_TABLE_OPTIONS
        )

        table.columns[0].justify = 'right'

        for node in self.pdfalyzer.stream_nodes():
            table.add_row(size_in_bytes_text(node.stream_length), node.__rich__())

        return table

    def _metadata_table(self) -> Table:
        """Build a table of metadata extracted from/computed about the PDF."""
        table = Table(
            header_style='bold',
            title=' Metadata',
            title_style='grey',
            title_justify='left',
            **DEFAULT_TABLE_OPTIONS
        )

        table.add_column('Property', justify='right')
        table.add_column('Value', min_width=40)

        if (metadata := self.pdfalyzer.pdf_reader.metadata or {}):
            for k, v in metadata.items():
                v = highlight(str(v)) if isinstance(v, str) and any(hif in v for hif in HIGHLIGHT_IF) else str(v)
                table.add_row(Text(k, style='wheat4'), v)
        else:
            table.add_row('', Text('(no formal metadata found in file)'), style='dim italic')

        def add_count_row(count_type: str, count_fxn: Callable[[], int]) -> None:
            row_label = count_txt(count_type)

            try:
                table.add_row(row_label, highlight(f"{count_fxn():,}"))
            except Exception as e:
                fail_msg = failed_count_msg(count_type)
                log.error(f"{fail_msg} {e}")
                table.add_row(row_label, Text(fail_msg, style='bright_red'))

        add_count_row('page', lambda: len(self.pdfalyzer.pdf_reader.pages))
        add_count_row('images', lambda: sum([len(p.images) for p in self.pdfalyzer.pdf_reader.pages]))
        add_count_row('revision', lambda: self.pdfalyzer.max_generation)
        return table



"""
PdfalyzerConfig object holds the unification of configuration options parsed from the command line
as well as those set by environment variables and/or a .pdfalyzer file.
"""
import sys
from argparse import Namespace
from os import path
from pathlib import Path
from typing import Callable, TypeVar

from yaralyzer.config import YaralyzerConfig
from yaralyzer.util.argument_parser import rules, tuning
from yaralyzer.util.constants import MAX_FILENAME_LENGTH
from yaralyzer.util.exceptions import print_fatal_error_and_exit
from yaralyzer.util.logging import log

from pdfalyzer.output.theme import COMPLETE_THEME_DICT, _debug_themes
from pdfalyzer.util.constants import PDFALYZE, PDFALYZER_UPPER
from pdfalyzer.util.helpers.filesystem_helper import DEFAULT_PDF_PARSER_PATH, PDF_PARSER_PATH_ENV_VAR
from pdfalyzer.util.output_section import ALL_STREAMS

T = TypeVar('T')

# These options will be read from env vars prefixed with YARALYZER, not PDFALYZER
# TODO: for some reasonm PDFALYZER_PATTERNS_LABEL and PDFALYZER_REGEX_MODIFIER are showing up
YARALYZER_SPECIFIC_OPTIONS = [
    action.dest
    for argument_group in [rules, tuning]
    for action in argument_group._group_actions
]


class PdfalyzerConfig(YaralyzerConfig):
    """Handles parsing of command line args and environment variables for Pdfalyzer."""

    # Override the class vars of same name in YaralyzerConfig
    ENV_VAR_PREFIX = PDFALYZER_UPPER
    COLOR_THEME = COMPLETE_THEME_DICT
    ONLY_CLI_ARGS = YaralyzerConfig.ONLY_CLI_ARGS + ['extract_binary_streams']

    pdf_parser_path: Path | None = None

    @classmethod
    def get_export_basepath(cls, export_method: Callable) -> str:
        """Build the path to an output file - everything but the extension"""
        export_type = export_method.__name__.removeprefix('print_')
        export_basename = f"{cls.args._export_basename}.{export_type}"

        if export_type == 'streams_analysis':
            if cls.args.streams != ALL_STREAMS:
                export_basename += f"_streamid{cls.args.streams}"

            export_basename += f"_maxdecode{YaralyzerConfig.args.max_decode_length}"

            if cls.args.extract_quoteds:
                export_basename += f"_extractquoteds-{','.join(cls.args.extract_quoteds)}"
            if cls.args.suppress_boms:
                export_basename += '_noBOMs'

        # YARA rules suffixes
        if cls.args.yara_rules_files:
            export_basename += f"__scannedby_" + ','.join(sorted([f.name for f in cls.args.yara_rules_files]))

            if cls.args.no_default_yara_rules:
                export_basename += '_customrulesonly'

        export_basename += cls.args.file_suffix

        if not cls.args.no_timestamps:
            export_basename += f"___{PDFALYZE}d_{cls.args._invoked_at_str}"

        max_filename_length = MAX_FILENAME_LENGTH - len(str(cls.args.output_dir.resolve()))
        return path.join(cls.args.output_dir, export_basename[:max_filename_length])

    @classmethod
    def _parse_arguments(cls) -> Namespace:
        """Overloads/extends YaralyzerConfig method of the same name."""
        if '--show-colors' in sys.argv and '--debug' in sys.argv:
            _debug_themes()

        args = super()._parse_arguments()
        args.extract_quoteds = args.extract_quoteds or []
        args._yaralyzer_standalone_mode = False  # TODO: this sucks
        args._export_basename = f"{args.file_prefix}{args.file_to_scan_path.name}"

        if not args.streams:
            if args.extract_quoteds:
                log.warning("--extract-quoted does nothing if --streams is not selected")
            if args.suppress_boms:
                log.warning("--suppress-boms has nothing to suppress if --streams is not selected")

        if args.no_default_yara_rules and not any(getattr(args, opt.dest) for opt in rules._group_actions):
            print_fatal_error_and_exit("--no-default-yara-rules requires at least one YARA rule argument")

        return args

    @classmethod
    def prefixed_env_var(cls, var: str) -> str:
        """Turns 'LOG_DIR' into 'PDFALYZER_LOG_DIR' etc. Overloads superclass method."""
        prefix = super().ENV_VAR_PREFIX if var in YARALYZER_SPECIFIC_OPTIONS else cls.ENV_VAR_PREFIX
        return (var if var.startswith(prefix) else f"{prefix}_{var}").upper()

    @classmethod
    def _set_class_vars_from_env(cls) -> None:
        """Set log related class vars and find path to pdf-parser.py (if any)."""
        super()._set_class_vars_from_env()
        cls.pdf_parser_path = cls.get_env_value(PDF_PARSER_PATH_ENV_VAR, Path) or DEFAULT_PDF_PARSER_PATH

        if not cls.pdf_parser_path.exists():
            log.warning(f"Configured PDF_PARSER_PATH is '{cls.pdf_parser_path}' but that file doesn't exist!")
            cls.pdf_parser_path = None



"""
`PdfTreeNode` decorates a `PdfObject` with tree structure information.
"""
from dataclasses import dataclass, field
from typing import Callable, Self

from anytree import NodeMixin, SymlinkNode
from pypdf.errors import PdfReadError
from pypdf.generic import ArrayObject, IndirectObject, PdfObject, StreamObject
from rich.markup import escape
from rich.table import Table
from rich.text import Text
from yaralyzer.output.console import console
from yaralyzer.output.theme import GREY_ADDRESS
from yaralyzer.util.helpers.rich_helper import DEFAULT_TABLE_OPTIONS

from pdfalyzer.decorators.pdf_object_properties import PdfObjectProperties
from pdfalyzer.output.tables.pdf_node_rich_table import get_stream_preview_rows
from pdfalyzer.output.theme import get_class_style_italic
from pdfalyzer.pdf_object_relationship import PdfObjectRelationship
from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError
from pdfalyzer.util.helpers.pdf_object_helper import pypdf_class_name
from pdfalyzer.util.helpers.string_helper import is_prefixed_by_any, numbered_list
from pdfalyzer.util.logging import log, log_console

DEFAULT_MAX_ADDRESS_LENGTH = 90
DECODE_FAILURE_LEN = -1
MAX_REFS_TO_WARN = 10

SORT_KEYS = {
    TYPE: '/A',
    SUBTYPE: '/AA',
    NAME: '/AAA',
    TITLE: '/AAAA',
}


@dataclass
class PdfTreeNode(NodeMixin):
    """
    PDF node decorator - wraps actual PDF objects to make them `anytree` nodes.
    Also adds decorators/generators for Rich text representation.

    Child/parent relationships should be set using the `add_child()` and `set_parent()`
    methods and not set directly.
    TODO: this could be done better with anytree hooks.
    """
    pdf_object: PdfObjectProperties

    # Computed fields
    all_references_processed: bool = False
    known_to_parent_as: str | None = None
    non_tree_relationships: list[PdfObjectRelationship] = field(default_factory=list)
    stream_data: bytes | None = None
    stream_length: int = 0

    @property
    def idnum(self) -> int:
        return self.pdf_object.idnum

    @property
    def label(self) -> str:
        return self.pdf_object.label

    @property
    def obj(self) -> PdfObject:
        return self.pdf_object.obj

    @property
    def type(self) -> str | None:
        return self.pdf_object.type

    def __post_init__(self):
        if self.contains_stream():
            try:
                self.stream_data = self.obj.get_data()
                self.stream_length = len(self.stream_data)
            except (NotImplementedError, PdfReadError) as e:
                log_console.print_exception()
                msg = f"Failed to decode stream in {self}, won't be able to scan/check this stream: {e}"
                log.error(msg)
                self.stream_data = msg.encode()
                self.stream_length = DECODE_FAILURE_LEN

    @classmethod
    def from_reference(cls, ref: IndirectObject, address: str) -> Self:
        """Alternate constructor to Build a `PdfTreeNode` from an `IndirectObject`."""
        return cls(PdfObjectProperties.from_reference(ref, address))

    @classmethod
    def from_obj(cls, obj: PdfObject, address: str, idnum: int) -> Self:
        """Alternate constructor to Build a `PdfTreeNode` from an `IndirectObject`."""
        return cls(PdfObjectProperties(obj, address, idnum))

    def set_parent(self, parent: Self | None, force: bool = False) -> None:
        """Set the parent of this node."""
        if parent is None:
            return
        elif self.parent is not None and self.parent != parent and not force:
            # Some objs in Arrays have the array's parent in /Parent so we link through
            if isinstance(self.parent.obj, ArrayObject):
                log.warning(f"Parent of {self} is already {self.parent} but it's an array; inserting {parent} as grandparent")
                parent.set_parent(self.parent)
                self.parent = parent
            else:
                log.warning(f"Cannot set {parent} as parent of {self}, parent is already {self.parent}")

            return

        self.parent = parent
        self.remove_non_tree_relationship(parent)
        self.known_to_parent_as = self.address_of_this_node_in_other(parent) or self.pdf_object.address
        log.info(f"  Added {parent} as parent of {self}" + (' by force' if force else ''))

    def add_child(self, child: Self) -> None:
        """Add a child to this node."""
        if next((c for c in self.children if c.idnum == child.idnum), None) is not None:
            log.debug(f"{child} is already child of {self}")
        else:
            child.set_parent(self)

    def add_non_tree_relationship(self, relationship: PdfObjectRelationship) -> None:
        """Add a relationship that points at this node's PDF object. TODO: doesn't include parent/child."""
        if relationship in self.non_tree_relationships:
            return

        self.non_tree_relationships.append(relationship)
        log.info(f'Added other relationship: {relationship} {self}')

    def remove_non_tree_relationship(self, from_node: Self) -> None:
        """Remove all non_tree_relationships from 'from_node' to this node."""
        relationships_to_remove = [r for r in self.non_tree_relationships if r.from_node == from_node]
        num_to_remove = len(relationships_to_remove)

        if num_to_remove > 1 and not all(r.reference_key in [FIRST, LAST] for r in relationships_to_remove):
            log.warning(f"Removing {num_to_remove} non-tree relationships between {from_node} and {self}.\n"
                        + numbered_list(relationships_to_remove))

        for relationship in relationships_to_remove:
            log.debug(f"Removing relationship {relationship} from {self}")
            self.non_tree_relationships.remove(relationship)

    def nodes_with_here_references(self) -> list[Self]:
        """Return a list of nodes that contain this node's PDF object as an IndirectObject reference."""
        return [r.from_node for r in self.non_tree_relationships if r.from_node]

    def non_tree_relationship_count(self) -> int:
        """Number of non parent/child relationships containing this node."""
        return len(self.non_tree_relationships)

    def unique_addresses(self) -> list[str]:
        """All the addresses in other nodes that refer to this object."""
        addresses = set([r.address for r in self.non_tree_relationships])

        if self.known_to_parent_as is not None:
            addresses.add(self.known_to_parent_as)

        return list(addresses)

    def references_to_other_nodes(self) -> list[PdfObjectRelationship]:
        """Returns all nodes referenced from node.obj (see PdfObjectRelationship definition)."""
        return PdfObjectRelationship.build_node_references(from_node=self)

    def contains_stream(self) -> bool:
        """Returns True for ContentStream, DecodedStream, and EncodedStream objects."""
        return isinstance(self.obj, StreamObject)

    def tree_address(self, max_length: int = DEFAULT_MAX_ADDRESS_LENGTH) -> str:
        """Creates a string like '/Catalog/Pages/Resources[2]/Font' truncated to max_length (if given)."""
        if self.label == TRAILER:
            return '/'
        elif self.parent is None:
            raise PdfWalkError(f"{self} does not have a parent; cannot get accurate node address.")
        elif self.parent.label == TRAILER:
            return self.known_to_parent_as   # noqa: if there's a parent there's always a known_to_parent_as

        address = self.parent.tree_address() + self.known_to_parent_as

        if max_length is None or max_length > len(address):
            return address

        return '...' + address[-max_length:][3:]

    def address_of_this_node_in_other(self, from_node: Self) -> str | None:
        """Find the local address used in 'from_node' to refer to this node."""
        refs_to_this_node = [
            ref for ref in from_node.references_to_other_nodes()
            if ref.to_obj.idnum == self.idnum
        ]

        if len(refs_to_this_node) == 1:
            return refs_to_this_node[0].address
        elif len(refs_to_this_node) == 0:
            # TODO: Hack city. /XRef streams are basically trailer nodes without any direct reference
            if self.parent and self.parent.label == TRAILER and self.type == XREF and XREF_STREAM in self.parent.obj:
                return XREF_STREAM
            elif self.label not in NON_STANDARD_ADDRESS_NODES:
                log.info(f"Could not find expected reference from {from_node} to {self}")
            else:
                return None
        else:
            address = refs_to_this_node[0].address

            # If other node's label doesn't start with a NON_STANDARD_ADDRESS string
            # AND any of the relationships pointing at this node use something other than a
            #     NON_STANDARD_ADDRESS_NODES string to refer here,
            # then print a warning about multiple refs.
            if not (is_prefixed_by_any(from_node.label, NON_STANDARD_ADDRESS_NODES)
                    or all(ref.address in NON_STANDARD_ADDRESS_NODES for ref in refs_to_this_node)):
                ref_addresses = [r.address for r in refs_to_this_node]
                common_key = next((k for k in MULT_REF_RESOURCE_KEYS if all(k in a for a in ref_addresses)), None)
                msg = f"Found {len(refs_to_this_node)} refs from {from_node} to {self}"
                refs_to_print = refs_to_this_node

                if common_key and len(ref_addresses) > MAX_REFS_TO_WARN:
                    msg += f", they're all {common_key} so showing only first {MAX_REFS_TO_WARN}"
                    refs_to_print = refs_to_this_node[0:MAX_REFS_TO_WARN]

                log.warning(f"{msg}:\n{numbered_list(refs_to_print)}\nCommon address of refs: {address}")

            return address

    def tree_relationships(self) -> list[Self]:
        """Returns parents and children."""
        return list(self.children) + ([self.parent] if self.parent is not None else [])

    def symlink_non_tree_relationships(self) -> None:
        """Create SymlinkNodes for this node's non parent/child (non-tree) relationships."""
        log.info(f"Symlinking {self}'s {self.non_tree_relationship_count()} other relationships...")

        for relationship in self.non_tree_relationships:
            if relationship.from_node in self.tree_relationships():
                log.info(f"{relationship} is non-tree symlink but {relationship.from_node} is now a real " \
                         f"parent or child of {self}, removing non-tree relationships")
                self.remove_non_tree_relationship(relationship.from_node)
            else:
                log.debug(f"   SymLinking {relationship} to {self}")
                SymlinkNode(self, parent=relationship.from_node)

    def descendants_count(self) -> int:
        """Count nodes in the tree that are children/grandchildren/great grandchildren/etc of this one."""
        return len(self.children) + sum([child.descendants_count() for child in self.children])

    def unique_labels_of_referring_nodes(self) -> list[str]:
        """Unique label strings of nodes referring here outside the parent/child hierarchy."""
        return list(set([r.from_node.label for r in self.non_tree_relationships]))

    def print_non_tree_relationships(self) -> None:
        """console.print this node's non tree relationships (represented by SymlinkNodes in the tree)."""
        self._write_non_tree_relationships(console.print)

    def log_non_tree_relationships(self) -> None:
        """log this node's non tree relationships (represented by SymlinkNodes in the tree)."""
        self._write_non_tree_relationships(log.info)

    # Presentation
    def as_tree_node_table(self, pdfalyzer: 'Pdfalyzer') -> Table:
        """
        Generate a Rich table representation of this node's PDF object and its properties.
        Table cols are [title, address, class name] (not exactly headers but sort of).
        Dangerous things like /JavaScript, /OpenAction, Type1 fonts, etc, will be highlighted red.

        Args:
            pdfalyzer (Pdfalyzer): Used to lookup nodes with properly assigned types to use in the table
        """
        title = f"{self.idnum}.{escape(self.label)}"
        address = escape(self.tree_address())

        if self.type == OBJ_STM:
            address += ' (should have been decompressed into other objs)'
            table_style = 'dim'
        else:
            table_style = ''

        table = Table(title, address, pypdf_class_name(self.obj), style=table_style, **DEFAULT_TABLE_OPTIONS)
        table.columns[0].header_style = f'reverse {self.pdf_object.label_style}'
        table.columns[1].header_style = 'dim'
        table.columns[1].overflow = 'fold'
        table.columns[2].header_style = get_class_style_italic(self.obj)

        if self.label != self.known_to_parent_as and self.type != TRAILER and self.known_to_parent_as:
            table.add_row(Text('AddressInParent', style='italic'), Text(self.known_to_parent_as), '', style='gray58')

        if isinstance(self.obj, dict):
            for k in sorted(self.obj.keys(), key=lambda k: SORT_KEYS.get(k, k)):
                row = self.pdf_object.get_table_row(k, pdfalyzer)

                # Make dangerous stuff look dangerous
                if (k in DANGEROUS_PDF_KEYS) or (self.label == FONT and k == SUBTYPE and v == TYPE1_FONT):
                    table.add_row(*[col.plain for col in row], style='bold reverse red')
                else:
                    table.add_row(*row)
        elif isinstance(self.obj, list):
            for i in range(len(self.obj)):
                table.add_row(*self.pdf_object.get_table_row(i, pdfalyzer))
        else:
            # Then it's a single element node like a URI, TextString, etc.
            table.add_row(*self.pdf_object.get_table_row(None, pdfalyzer, empty_3rd_col=True))

        for row in get_stream_preview_rows(self):
            table.add_row(*(row + [Text('')]))

        return table

    def _colored_address(self, max_length: int = DEFAULT_MAX_ADDRESS_LENGTH) -> Text:
        """Rich text version of tree_address()."""
        text = Text('@', style='bright_white')
        return text.append(self.tree_address(max_length), style=GREY_ADDRESS)

    def _write_non_tree_relationships(self, write_method: Callable) -> None:
        """Use write_method() to write self.non_tree_relationships."""
        write_method(f"{escape(str(self))} parent from candidates:")

        for i, r in enumerate(self.non_tree_relationships):
            write_method(f"  {i + 1}. {escape(str(r))}, Descendant Count: {r.from_node.descendants_count()}")

    def __repr__(self) -> str:
        return self.__str__()

    def __rich__(self) -> Text:
        """Like the __rich__() of PdfObjectProperties but with address appended."""
        return self.pdf_object.__rich__()[:-1] + self._colored_address() + Text('>')

    def __str__(self) -> str:
        return self.pdf_object.__rich__().plain
